#!/usr/bin/env python
""" 
Collection of functions to trim audio signals.
"""
#
# Authors:  Juan Sebastian ULLOA <lisofomia@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

# =============================================================================
# Load the modules
# =============================================================================
# Import external modules
import numpy as np

#%%
# =============================================================================
# public functions
# =============================================================================
def wave2frames (s, Nt=512):
    """
    Reshape a sound waveform (ie vector) into a serie of frames (ie matrix) of
    length Nt.
    
    Parameters
    ----------
    s : 1d ndarray of floats (already divided by the number of bits)
        Vector containing the sound waveform 
    Nt : int, optional, default is 512
        Number of points per frame
                
    Returns
    -------
    timeframes : 2d ndarray of floats
        Matrix containing K frames (row) with Nt points (column), K*N <= length (s)
        
    Examples
    --------
    
    >>> import numpy as np  
    
    Fast method to estimate the envelope of a waveform
    
    >>> s, fs = maad.sound.load('../data/rock_savanna.wav')
    >>> frames = maad.sound.wave2frames(s, Nt=32)
    >>> env_fast = np.max(abs(frames),0) 
    
    Comparison with the classic method with Hilbert transform 
    
    >>> env_hilbert = maad.sound.envelope(s, mode='hilbert')
    
    Compute the time vector for the vector wave.

    >>> t = np.arange(0,len(s),1)/fs
    
    Compute the time vector for the vector env_fast.
    
    >>> t_env_fast = np.arange(0,len(env_fast),1)*len(s)/fs/len(env_fast)
    
    Plot 0.1s of the envelope and 0.1s of the abs(s).
    
    >>> import matplotlib.pyplot as plt
    >>> fig1, ax1 = plt.subplots(figsize=(10,4))
    >>> ax1.plot(t[t<0.1], abs(s[t<0.1]), label='abs(s)', lw=0.7)
    >>> ax1.plot(t[t<0.1], env_hilbert[t<0.1], label='env(s) - hilbert option', lw=0.7)
    >>> ax1.plot(t_env_fast[t_env_fast<0.1], env_fast[t_env_fast<0.1], label='env(s) - fast option', lw=0.7)
    >>> ax1.set_xlabel('Time [sec]')
    >>> ax1.set_ylabel('Amplitude')
    >>> ax1.legend()
    
    """
    # transform wave into array
    s = np.asarray(s)
    # compute the number of frames
    K = len(s)//Nt
    # Reshape the waveform (ie vector) into a serie of frames (ie 2D matrix)
    timeframes = s[0:K*Nt].reshape(-1,Nt).transpose()
    
    return timeframes




# -*- coding: utf-8 -*-
"""
Segmentation methods
====================

The module ``rois`` has a collection of functions to segment and find regions of interest in audio and spectrograms.

Temporal (1D)
-------------
.. autosummary::
    :toctree: generated/

    find_rois_cwt
    
Spectro-temporal (2D)
---------------------
.. autosummary::
    :toctree: generated/
    
    create_mask
    select_rois
    rois_to_imblobs
    template_matching
    spectrogram_local_max

"""

from .rois_2d import (create_mask,
                     select_rois,
                     rois_to_imblobs,
                     spectrogram_local_max)

from .rois_1d import (find_rois_cwt)

from .template_matching_func import (template_matching)

__all__ = [ # rois 2d
           'create_mask',
           'select_rois',
           'rois_to_imblobs',
           'spectrogram_local_max',
           # rois 1d
           'find_rois_cwt',
           # template matching
           'template_matching']



#!/usr/bin/env python
""" Utilitary functions to parse and read audio and text files. """
#
# Authors:  Juan Sebastian ULLOA <lisofomia@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

#%%
# =============================================================================
# Load the modules
# =============================================================================
# Import external modules
import numpy as np 
import pandas as pd
import re
import os
import glob
from datetime import datetime
from pathlib import Path # in order to be Windows/linux/MacOS compatible


#%%
# =============================================================================
# Private functions
# =============================================================================
def _date_from_filename (filename):
    """
    Extract date and time from the filename. Return a datetime object
    
    Parameters
    ----------
    filename : string
    The filename must follow this format :
    XXXX_yyyymmdd_hhmmss.wav
    with yyyy : year / mm : month / dd: day / hh : hour (24hours) /
    mm : minutes / ss : seconds
            
    Returns
    -------
    date : object datetime
        This object contains the date of creation of the file extracted from
        the filename postfix. 
    """
    # date by default
    date = datetime(1900,1,1,0,0,0,0)
    # test if it is possible to extract the recording date from the filename
    if filename[9:13].isdigit(): 
        yy=int(filename[9:13])
    else:
        return date
    if filename[13:15].isdigit(): 
        mm=int(filename[13:15])
    else:
        return date
    if filename[15:17].isdigit(): 
        dd=int(filename[15:17])
    else:
        return date
    if filename[18:20].isdigit(): 
        HH=int(filename[18:20])
    else:
        return date
    if filename[20:22].isdigit(): 
        MM=int(filename[20:22])
    else:
        return date
    if filename[22:24].isdigit(): 
        SS=int(filename[22:24])
    else:
        return date

    # extract date and time from the filename
    date = datetime(year=yy, month=mm, day=dd, hour=HH, minute=MM, second=SS, 
                    microsecond=0)
    
    return date


#%%
# =============================================================================
# Public functions
# =============================================================================
def read_audacity_annot (audacity_filename):
    """
    Read Audacity annotations file (or labeling file) and return a Pandas Dataframe
    with the bounding box and the label of each region of interest (ROI). Allows to
    read annotations with standard Audacity style (temporal selection) and with
    spectral selection style (spectro-temporal selection). If the file exists but has no 
    annotations, the function returns and empty dataframe.
    
    Parameters
    ----------
    audacity_filename : String
        Path to the audacity file

    Returns
    -------
    tab_out : Pandas Dataframe 
        Region of interest with time-frequency limits and manual annotation label
    
    References
    ----------
    https://manual.audacityteam.org/man/label_tracks.html  
    
    Examples
    --------
    >>> from maad import sound
    >>> from maad.util import power2dB, read_audacity_annot, format_features, overlay_rois
    >>> s, fs = sound.load('../data/cold_forest_daylight.wav')
    >>> Sxx_power, tn, fn, ext = sound.spectrogram(s, fs, nperseg=1024, noverlap=1024//2)
    >>> Sxx_db = power2dB(Sxx_power) + 96
    >>> df_rois = read_audacity_annot('../data/cold_forest_daylight_label.txt') 
    >>> df_rois = format_features(df_rois, tn, fn)
    >>> overlay_rois(Sxx_db, df_rois, **{'vmin':0,'vmax':96,'extent':ext})
    
    """
    # try to read file with tab delimiter (if the file is not empty)
    try:
        tab_in = pd.read_csv(audacity_filename, delimiter='\t', header=None)

        # test if time-frequency annotation (1st column contain '/')
        # Hack to force the type of the column to be string in order to test if 
        # the column contains a character
        tab_in[0] = tab_in[0].astype('str')
        if (tab_in[0].str.contains(r"\\", na = False).sum() > 0) :

            # arrange data
            t_info = tab_in.loc[np.arange(0, len(tab_in), 2), :]
            t_info = t_info.rename(index=str, columns={
                                0: 'min_t', 1: 'max_t', 2: 'label'})
            t_info = t_info.reset_index(drop=True)
    
            f_info = tab_in.loc[np.arange(1, len(tab_in)+1, 2), :]
            f_info = f_info.rename(index=str, columns={
                                0: 'slash', 1: 'min_f', 2: 'max_f'})
            f_info = f_info.reset_index(drop=True)
    
            # return dataframe
            tab_out = pd.concat(
                                [t_info['label'].astype('str'),
                                t_info['min_t'].astype('float32'),
                                f_info['min_f'].astype('float32'),
                                t_info['max_t'].astype('float32'),
                                f_info['max_f'].astype('float32')],  axis=1)
        else :
            tab_in = tab_in.rename(index=str, columns={
                                0: 'min_t', 1: 'max_t', 2: 'label'})
            tab_in['min_f'] = np.nan
            tab_in['max_f'] = np.nan
            
            # return dataframe
            tab_out = pd.concat([
                                tab_in['label'].astype('str'),
                                tab_in['min_t'].astype('float32'),
                                tab_in['min_f'].astype('float32'),
                                tab_in['max_t'].astype('float32'),
                                tab_in['max_f'].astype('float32')],  axis=1)
    except :
        tab_out = pd.DataFrame()

    return tab_out

#%%
def write_audacity_annot(fname, df_rois, save_file=True):
    """ 
    Write audio segmentation to text file in Audacity format, a file that can be imported
    and modified with Audacity. If the dataframe has no frequency delimiters, annotations
    are saved with standard Audacity format (temporal segmentation). If the dataframe has
    temporal and frequencial delimiters, the annotations are saved as spectral selection 
    style (spectro-temporal selection). If the dataframe is empty, the function saves an 
    empty file.
    
    Parameters
    ----------
    fname: str
        filename to save the segmentation
    df_rois: pandas dataframe
        Dataframe containing the coordinates corresponding to sound signatures
        In case of only temporal annotations : df_rois must contain at least
        the columns 'mint_t', 'max_t' 
        In case of bounding box (temporal eand frequency limits) :: df_rois 
        must contain at least the columns 'min_t', 'max_t', 'min_f', 'max_f'
    save_file: bool, optional, default=True
        If True, the file is saved. If False, the file is not saved.

    Returns
    -------
    df_to_save
        Dataframe that has been saved
    
    Examples
    --------
    >>> s, fs = maad.sound.load('../data/cold_forest_daylight.wav')
    >>> Sxx_power, tn, fn, ext = maad.sound.spectrogram(s, fs)
    >>> Sxx_db = maad.util.power2dB(Sxx_power) + 96
    >>> Sxx_power_noNoise= maad.sound.median_equalizer(Sxx_power)
    >>> Sxx_db_noNoise = maad.util.power2dB(Sxx_power_noNoise)
    >>> Sxx_db_noNoise_smooth = maad.sound.smooth(Sxx_db_noNoise, std=0.5)
    >>> im_mask = maad.rois.create_mask(im=Sxx_db_noNoise_smooth, mode_bin ='relative', 
                                        bin_std=8, bin_per=0.5)
    >>> im_rois, df_rois = maad.rois.select_rois(im_mask, min_roi=25, max_roi=None)
    >>> df_rois = maad.util.format_features(df_rois, tn, fn)
    
    Change path to save the file containing the labels position
    
    >>> df_to_save = maad.util.write_audacity_annot('save.txt', df_rois)
    
    Import the wav file then the label file in Audacity
    
    """
    if df_rois.empty:  # empty DataFrame
        print(f'{fname} > No detection found')
        df_to_save = pd.DataFrame(data=None)
        
    else:  
        # if there is no label, create a vector with incremental values
        if 'label' not in df_rois:
            df_rois['label'] = np.arange(0,len(df_rois))
        
        # if no frequency coordinates, only temporal annotations
        if ('min_f' not in df_rois) or ('max_f' not in df_rois):
            df_to_save = pd.DataFrame({'min_t':df_rois.min_t, 
                                    'max_t':df_rois.max_t, 
                                    'label':df_rois.label})
        else:
            df_to_save_odd = pd.DataFrame({'index': np.arange(0,len(df_rois)*2,2),
                                        'min_t':df_rois.min_t, 
                                        'max_t':df_rois.max_t, 
                                        'label':df_rois.label})
            df_to_save_even = pd.DataFrame({'index': np.arange(1,len(df_rois)*2,2),
                                        'min_t':'\\', 
                                        'max_t':df_rois.min_f, 
                                        'label':df_rois.max_f})
            df_to_save = pd.concat([df_to_save_odd,df_to_save_even])
            df_to_save = df_to_save.set_index('index')
            df_to_save = df_to_save.sort_index()
            
    if save_file:
        df_to_save.to_csv(fname, index=False, header=False, sep='\t') 
    else:
        pass
    
    return df_to_save

#%% 
def read_raven_annot(raven_filename):
    """
    Read Raven annotations file (or labeling file) and return a Pandas Dataframe
    with the bounding box and the label of each region of interest (ROI). If the file 
    exists but has no annotations, the function returns and empty dataframe.
    
    Parameters
    ----------
    raven_filename : string
        Path to the annotation file

    Returns
    -------
    tab_out : Pandas Dataframe 
        Region of interest with time-frequency limits and manual annotation label
    
    References
    ----------
    http://ravensoundsoftware.com/wp-content/uploads/2017/11/Raven14UsersManual.pdf
        
    """
    df_out = pd.read_csv(raven_filename, sep='\t')
    return df_out
    
#%%
def write_raven_annot(fname, df_rois, save_file=True):
    """ 
    Write audio segmentation to text file in Raven format, a file that can be imported
    and modified with Raven. If the dataframe has no frequency delimiters, annotations
    are saved with standard Audacity format (temporal segmentation). If the dataframe has
    temporal and frequencial delimiters, the annotations are saved as spectral selection 
    style (spectro-temporal selection). If the dataframe is empty, the function saves an 
    empty file.
    
    Parameters
    ----------
    fname: str
        filename to save the segmentation
    df_rois: pandas dataframe
        Dataframe containing the coordinates corresponding to sound signatures
        For bounding box (temporal eand frequency limits) :: df_rois 
        must contain at least the columns 'min_t', 'max_t', 'min_f', 'max_f'
            
    Returns
    -------
    df_out: pandas dataframe
        Dataframe that has been saved in Raven format
    
    Examples
    --------
    >>> from maad import sound, rois, util
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> df_rois = rois.find_rois_cwt(s, fs, flims=(3000,8000), tlen=2, th=0)
    >>> df_rois['Label'] = 'Spinetail'
    >>> df_raven = util.write_raven_annot('spinetail_annotations.txt', df_rois)
    
    """
    df_out = df_rois.copy()
    # Save empty file if dataframe is empty
    if df_out.size==0:
        print(fname, '> No detection found')
        df = pd.DataFrame(data=None)
        df.to_csv(fname, sep='\t', header=False, index=False)
        
    else:
        # Format dataframe and save
        # add basic raven columns if needed
        if not('Selection' in df_out.columns):
            df_out['Selection'] = np.arange(1, len(df_out)+1)
        
        if not('View' in df_out.columns):
            df_out['View'] = 'Spectrogram 1'
        
        if not('Channel' in df_out.columns):
            df_out['Channel'] = 1
        
        # change column names
        df_out.rename(columns={
                        'min_t': 'Begin Time (s)', 
                        'max_t': 'End Time (s)',
                        'min_f': 'Low Freq (Hz)',
                        'max_f': 'High Freq (Hz)'}, inplace=True)
        
        # reorder column names
        colname_raven = ['Selection', 'View', 'Channel', 'Begin Time (s)',
                        'End Time (s)', 'Low Freq (Hz)', 'High Freq (Hz)']
        colname_order = colname_raven + df_out.columns[~df_out.columns.isin(colname_raven)].tolist()
        df_out = df_out.reindex(columns=colname_order)
        
        if save_file:
            df_out.to_csv(fname, sep='\t', index=False)
        else:
            pass
    
    return df_out

#%%
def date_parser(datadir, dateformat='%Y%m%d_%H%M%S', extension='.wav', prefix = '', verbose=False):
    """
    Extracts dates from filenames in a given folder and subfolders.

    Parameters
    ----------
    datadir : str
        Path to the folder to search for files.
    dateformat : str, optional
        Format string specifying the datetime pattern to extract.
        The default is'%Y%m%d_%H%M%S'
        For more information about the format codes, refer to the
        `strftime format documentation <https://strftime.org/>`_.
    extension : str, optional, 
        File extension to filter files by (e.g., '.wav', '.mp3').
        The default is '.wav'.
    prefix : str, optional, 
        Prefix of the filenames to match.
        The default is ''.
    verbose : bool, optional
        If True, print the filenames as they are processed.
        The default is False.

    Returns
    -------
    pandas.DataFrame
        DataFrame containing the extracted dates as the index 'Date',
        and the full file paths in a 'file' column.

    Raises
    ------
    ValueError
        If the datetime_format is invalid or does not match the filenames.

    Notes
    -----
    This function searches for files in the specified folder and its subfolders
    that have the given extension and match the specified prefix. It extracts
    the dates from the filenames using the provided datetime_format.

    The extracted dates are set as the index of the resulting DataFrame. The
    'file' column contains the full file paths.

    Examples
    --------
    >>> folder_path = '../../data/indices/'
    >>> ext = '.wav'
    >>> datetime_format = '%Y%m%d_%H%M%S'
    >>> df = maad.util.date_parser(datadir=folder_path, dateformat=datetime_format, extension=ext)
    >>> df
                                                                    file
    Date	
    2019-05-22 00:00:00	../../data/indices/S4A03895_20190522_000000.wav
    2019-05-22 00:15:00	../../data/indices/S4A03895_20190522_001500.wav
    2019-05-22 00:30:00	../../data/indices/S4A03895_20190522_003000.wav
    2019-05-22 00:45:00	../../data/indices/S4A03895_20190522_004500.wav
    2019-05-22 01:00:00	../../data/indices/S4A03895_20190522_010000.wav
                    ...	                                            ...
    2019-05-22 22:45:00	../../data/indices/S4A03895_20190522_224500.wav
    2019-05-22 23:00:00	../../data/indices/S4A03895_20190522_230000.wav
    2019-05-22 23:15:00	../../data/indices/S4A03895_20190522_231500.wav
    2019-05-22 23:30:00	../../data/indices/S4A03895_20190522_233000.wav
    2019-05-22 23:45:00	../../data/indices/S4A03895_20190522_234500.wav


    >>> df = maad.util.date_parser("../../data/indices/", dateformat='SM4', verbose=False)
    >>> list(df)
    >>> df
                                                                    file
    Date	
    2019-05-22 00:00:00	../../data/indices/S4A03895_20190522_000000.wav
    2019-05-22 00:15:00	../../data/indices/S4A03895_20190522_001500.wav
    2019-05-22 00:30:00	../../data/indices/S4A03895_20190522_003000.wav
    2019-05-22 00:45:00	../../data/indices/S4A03895_20190522_004500.wav
    2019-05-22 01:00:00	../../data/indices/S4A03895_20190522_010000.wav
                    ...	                                            ...
    2019-05-22 22:45:00	../../data/indices/S4A03895_20190522_224500.wav
    2019-05-22 23:00:00	../../data/indices/S4A03895_20190522_230000.wav
    2019-05-22 23:15:00	../../data/indices/S4A03895_20190522_231500.wav
    2019-05-22 23:30:00	../../data/indices/S4A03895_20190522_233000.wav
    2019-05-22 23:45:00	../../data/indices/S4A03895_20190522_234500.wav


    """    

    file_pattern = os.path.join(datadir, f'**/{prefix}*{extension}')
    file_list = glob.glob(file_pattern, recursive=True)
    data = []

    for file_path in file_list:

        # Extract the filename from the full path
        filename = os.path.basename(file_path)

        if dateformat == 'SM4':
            date =_date_from_filename(filename)
            data.append({'Date': date, 'file': file_path})
        
        elif dateformat == 'POSIX':
            posix_time = int(Path(filename).stem, 16)
            date = datetime.utcfromtimestamp(posix_time).strftime('%Y-%m-%d %H:%M:%S')
            data.append({'Date': date, 'file': file_path})

        else: 
            # Construct a regex pattern to extract the date from the filename
            pattern = _construct_pattern(dateformat)
            # Search for the date in the filename
            match = pattern.search(filename)
            # If a match is found, extract the date
            if match:
                if verbose:
                    print(f'File: {filename}')
                # Extract the date from the filename
                date_str = match.group()
                # Parse the date string
                try:
                    date = datetime.strptime(date_str, dateformat)
                    data.append({'Date': date, 'file': file_path})
                except ValueError:
                    print(f"Error parsing date: {date_str} in file: {filename}. The default date and time 1900-01-01 00:00:00 will be used.")
                    # date by default
                    data.append({'Date': "1900-01-01 00:00:01", 'file': file_path})
            else:
                print(f"No date found in file: {file_path}. The default date and time 1900-01-01 00:00:00 will be used.")
                # date by default 1900-01-01 00:00:00
                data.append({'Date': "1900-01-01 00:00:01", 'file': file_path})
    if len(data) > 0:
        df = pd.DataFrame(data)
        df.set_index('Date', inplace=True)
        # convert index to datetime 
        df.index = pd.DatetimeIndex(df.index)
        # sort dataframe by date
        df = df.sort_index(axis=0)
    else:
        df = pd.DataFrame()

    return df

def _construct_pattern(datetime_format):
    format_dict = {
        '%Y': r'(\d{4})',
        '%y': r'(\d{2})',
        '%m': r'(0[1-9]|1[0-2])',
        '%d': r'(0[1-9]|1\d|2[0-9]|3[01])',
        '%H': r'([01]\d|2[0-3])',
        '%I': r'(0[1-9]|1[0-2])',
        '%p': r'(AM|PM)',
        '%M': r'([0-5]\d)',
        '%S': r'([0-5]\d)',
        '%f': r'(\d{6})',
        '%j': r'(\d{3})',
        '%U': r'(\d{2})',
        '%W': r'(\d{2})',
        '%w': r'(\d)',
        '%A': r'(\w+)',
        '%a': r'(\w+)',
        '%B': r'(\w+)',
        '%b': r'(\w+)',
        '%c': r'(.+)',
        '%x': r'(.+)',
        '%X': r'(.+)',
        '%%': r'%',
    }
    pattern = datetime_format
    for code, regex in format_dict.items():
        pattern = pattern.replace(code, regex)
    pattern = re.compile(pattern)

    return pattern



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" S
Segmentation methods for 1D signals.
This module gathers a collection of functions to detect regions of interest (ROIs)
in the temporal domain.
"""
#
# Authors:  Juan Sebastian ULLOA <lisofomia@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

# =============================================================================
# Load the modules
# =============================================================================
# Import external modules
import numpy as np
import pywt
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import pandas as pd

# import internal modules
from maad.sound import sinc
from maad import sound, util

#%%
# =============================================================================
# Private functions
# =============================================================================
def _corresp_onset_offset(onset, offset, tmin, tmax):
    """ 
    Check that each onsets have a corresponding offset.

    Parameters
    ----------
    onset: ndarray
        array with onset from find_rois_1d
    offset: ndarray
        array with offset from find_rois_1d
    tmin: float
        Start time of wav file  (in s)
    tmax:
        End time of wav file  (in s)
    
    Return
    ------
    onset : ndarray
        onset with corresponding offset
    offset : ndarray
        offset with corresponding onset
    """
    if onset[0] > offset[0]:      # check start
        onset = np.insert(onset,0,tmin)
    else:
        pass
    if onset[-1] > offset[-1]:      # check end
        offset = np.append(offset,tmax)
    else:
        pass
    return onset, offset

#%%
def _energy_windowed(s, fs: float, wl: int=512):
    """ 
    Computse windowed energy on an audio signal.
    
    Computes the energy of the signals by windows of length wl. Used to amplify sectors where the density of energy is higher
    
    Parameters
    ----------
    s : ndarray
        input signal
    fs : float
        frequency sampling of the signal, used to keep track of temporal information of the signal
    wl : int, default is 512
        length of the window to summarize the rms value

    Returns
    -------
    time : ndarray
        temporal index vector
    s_rms : ndarray
        windowed rms signal
    """
    
    s_aux = np.pad(s, (0, wl-len(s)%wl), 'reflect')  # padding
    s_aux = s_aux**2 
    #  s_aux = np.abs(s_aux) # absolute value. alternative option
    s_aux = np.reshape(s_aux,(int(len(s_aux)/wl),wl))
    s_rms = np.mean(s_aux,1)
    time = np.arange(0,len(s_rms)) * wl / fs + wl*0.5/fs
    return time, s_rms

#%%
# =============================================================================
# Public functions
# =============================================================================
def find_rois_cwt(s, fs, flims, tlen, th: float=0, display=False, save_df=False, 
                  savefilename='rois.csv', **kwargs):
    """
    Find region of interest using known estimates of signal length and frequency limits.
    
    The general approach is based on continous wavelet transform following a three step process
    
    1. Filter the signal with a bandpass sinc filter
    
    2. Smoothing the signal by convolving it with a Mexican hat wavelet (Ricker wavelet) [1]
    
    3. Binarize the signal applying a linear threshold
        
    Parameters
    ----------
    s : ndarray
        input signal
    fs : float
        frequency sampling of the signal, used to keep track of temporal information of the signal
    flims : int
        upper and lower frequencies (in Hz) 
    tlen : int 
        temporal length of signal searched (in s)
    th : float, optional
        threshold to binarize the output
    display: boolean, optional, default is False
        plot results if set to True, default is False
    save_df : boolean, optional
        save results to csv file
    savefilename : str, optional
        Name of the file to save the table as comma separatd values (csv)        
    
    Returns
    -------
    rois : pandas DataFrame
        an table with temporal and frequencial limits of regions of interest            
    
    References
    ----------
    .. [1] Pan Du, Warren A. Kibbe, Simon M. Lin, Improved peak detection in mass spectrum by incorporating continuous wavelet transform-based pattern matching, Bioinformatics, Volume 22, Issue 17, 1 September 2006, Pages 2059–2065, `DOI: 10.1093/bioinformatics/btl355 <https://doi.org/10.1093/bioinformatics/btl355>`_ 
    
    Examples
    --------
    >>> from maad import sound, rois
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> rois.find_rois_cwt(s, fs, flims=(4500,8000), tlen=2, th=0, display=True)
        min_f     min_t   max_f     max_t
    0  4500.0   0.74304  8000.0   2.50776
    1  4500.0   5.10839  8000.0   7.33751
    2  4500.0  11.23846  8000.0  13.37469
    3  4500.0  16.16109  8000.0  18.29732
    """
    # filter signal
    s_filt = sinc(s, flims, fs, atten=80, transition_bw=0.8)
    # rms: calculate window of maximum 5% of tlen. improves speed of cwt
    wl = 2**np.floor(np.log2(tlen*fs*0.05)) 
    t, s_rms = _energy_windowed(s_filt, fs, int(wl))
    # find peaks
    cwt_width = [round(tlen*fs/wl/2)]
    npad = 5 ## seems to work with 3, but not sure
    s_rms = np.pad(s_rms, np.int64(cwt_width[0]*npad), 'reflect')  ## add pad
    s_cwt, _ = pywt.cwt(s_rms, cwt_width, 'mexh')
    s_cwt = s_cwt[0][np.int64(cwt_width[0]*npad):len(s_cwt[0])-np.int64(cwt_width[0]*npad)] ## rm pad
    # find onset and offset of sound
    segments_bin = np.array(s_cwt > th)
    onset = t[np.where(np.diff(segments_bin.astype(int)) > 0)]+t[0]  # there is delay because of the diff that needs to  be accounted
    offset = t[np.where(np.diff(segments_bin.astype(int)) < 0)]+t[0]
    # format for output
    if onset.size==0 or offset.size==0:
    # No detection found
        print('Warning: No detection found')
        df = pd.DataFrame(data=None)
        if save_df==True:
            df.to_csv(savefilename, sep=',',header=False, index=False)

    else:
    # A detection was found, save results to csv
        onset, offset = _corresp_onset_offset(onset, offset, tmin=0, tmax=len(s)/fs)
        rois_tf = np.transpose([np.repeat(flims[0],repeats=len(onset)),
                                np.round(onset,5),  
                                np.repeat(flims[1],repeats=len(onset)),
                                np.round(offset,5)])
        cols=['min_f', 'min_t','max_f', 'max_t']
        df = pd.DataFrame(data=rois_tf,columns=cols)
        if save_df==True:
            df.to_csv(savefilename, sep=',', header=True, index=False)

    # Display
    if display==True: 
        figsize = kwargs.pop('figsize',(12,6))
        cmap = kwargs.pop('cmap','gray')
        nfft = kwargs.pop('nperseg',512)
        noverlap = kwargs.pop('noverlap',512)
        nperseg = kwargs.pop('nperseg',1024)
        db_range = kwargs.pop('db_range',80)
        # plot wavelet
        fig,(ax1,ax2) = plt.subplots(2,1,figsize=figsize)
        ax1.margins(x=0)
        ax1.plot(s_cwt)
        ax1.set_xticks([])
        ax1.set_ylabel('Amplitude')
        ax1.grid(True)
        ax1.hlines(th, 0, len(s_cwt), linestyles='dashed', colors='r')
        #plot spectrogram
        Sxx, tn, fn, ext = sound.spectrogram(s, fs, nperseg = nperseg, noverlap=noverlap, mode='psd')
        util.plot_spectrogram(Sxx, ext, db_range, ax=ax2, colorbar=False)
        ax2.set_ylabel('Frequency (Hz)')
        ax2.set_xlabel('Time (s)')
        if not(df.empty):
            for idx, row in df.iterrows():
                xy = (row.min_t, row.min_f)
                width = row.max_t- row.min_t
                height = row.max_f- row.min_f
                rect = patches.Rectangle(xy, width, height, lw=1, 
                                        edgecolor='yellow', facecolor='none')
                ax2.add_patch(rect)
        plt.show()
    
    return df

if __name__ == "__main__":
    import doctest
    doctest.testmod()


# -*- coding: utf-8 -*-
"""
Utilities
=========

The module ``utils`` has a handful of useful set of tools used in the audio analysis framework.

Visualization
-------------
.. autosummary::
    :toctree: generated/
    
    rand_cmap
    crop_image
    save_figlist
    plot1d
    plot_wave
    plot_spectrum
    plot2d
    plot_spectrogram
    overlay_rois
    overlay_centroid
    plot_features_map
    plot_features
    heatmap_by_date_and_time
    plot_correlation_map
    plot_shape
    false_Color_Spectro

Mathematical
------------
.. autosummary::
    :toctree: generated/
    
    running_mean
    get_unimode
    entropy
    rms
    kurtosis
    skewness
    moments

Parser
------
.. autosummary::
    :toctree: generated/
    
    read_audacity_annot
    write_audacity_annot
    read_raven_annot
    write_raven_annot
    date_parser

Miscellaneous
-------------
.. autosummary::
    :toctree: generated/
    
    index_bw
    into_bins
    shift_bit_length
    rle
    linear_scale
    amplitude2dB
    power2dB
    dB2amplitude
    dB2power
    mean_dB
    add_dB
    nearest_idx
    get_df_single_row
    format_features
    crossfade
    crossfade_list

Xeno-Canto
----------
.. autosummary::
    :toctree: generated/
    
    xc_query
    xc_multi_query
    xc_selection
    xc_download

Audio metadata
--------------
.. autosummary::
    :toctree: generated/

    check_file_format
    audio_header
    filename_info
    get_metadata_file
    get_metadata_dir

"""

from .miscellaneous import (
                        index_bw,
                        into_bins,
                        shift_bit_length,
                        rle,
                        linear_scale,
                        amplitude2dB,
                        power2dB,
                        dB2amplitude,
                        dB2power,
                        mean_dB,
                        add_dB,
                        nearest_idx,
                        get_df_single_row,
                        format_features,
                        crossfade,
                        crossfade_list)

from .visualization import (
                        rand_cmap,
                        crop_image,
                        save_figlist,
                        plot1d,
                        plot_wave,
                        plot_spectrum,
                        plot2d,
                        plot_spectrogram,
                        overlay_rois,
                        overlay_centroid,
                        plot_features_map,
                        heatmap_by_date_and_time,
                        plot_features,
                        plot_correlation_map,
                        plot_shape,
                        false_Color_Spectro)

from .math_func import (
                        running_mean,
                        get_unimode,
                        entropy,
                        rms,
                        kurtosis,
                        skewness,
                        moments)                     

from .parser import (
                    read_audacity_annot,
                    write_audacity_annot,
                    read_raven_annot,
                    write_raven_annot,
                    date_parser)

from .xeno_canto import (
                        xc_query,
                        xc_multi_query,
                        xc_selection,
                        xc_download,
                        # xc_save_csv,
                        # xc_read_csv
                        )

from .audio_metadata_utilities import (
                                    check_file_format,
                                    audio_header,
                                    filename_info,
                                    get_metadata_file,
                                    get_metadata_dir)

__all__ = [
            # miscellaneous 
            'index_bw',
            'into_bins',
            'shift_bit_length',
            'rle',
            'linear_scale',
            'amplitude2dB',
            'power2dB',
            'dB2amplitude',
            'dB2power',
            'mean_dB',
            'add_dB',
            'nearest_idx',
            'get_df_single_row',
            'format_features',
            'crossfade',
            'crossfade_list',
            #  visualization      
            'rand_cmap',
            'crop_image',
            'save_figlist',
            'plot1d',
            'plot_wave',
            'plot_spectrum',
            'plot2d',
            'plot_spectrogram',
            'overlay_rois',
            'overlay_centroid',
            'plot_features_map',
            'plot_features',
            'heatmap_by_date_and_time',
            'plot_correlation_map',
            'plot_shape',
            'false_Color_Spectro',
            # math_func       
            'running_mean',
            'get_unimode',
            'entropy',
            'rms',
            'kurtosis',
            'skewness',
            'moments',
            # parser
            'read_audacity_annot',
            'write_audacity_annot',
            'read_raven_annot',
            'write_raven_annot',
            'date_parser',
            # xeno_canto
            'xc_query',
            'xc_multi_query',
            'xc_selection',
            'xc_download',
            # 'xc_save_csv',
            # 'xc_read_csv',
            # audio_metadata_utilities
            'check_file_format',
            'audio_header',
            'filename_info',
            'get_metadata_file',
            'get_metadata_dir'
            ]



# -*- coding: utf-8 -*-
"""
Sound pressure level
====================

The module ``spl`` is a collection of functions used to describe the physics of acoustic waves.


Conversion to sound pressure level (SPL)
----------------------------------------
.. autosummary::
    :toctree: generated/
    
    wav2volt
    volt2pressure
    wav2pressure
    pressure2dBSPL
    dBSPL2pressure
    power2dBSPL
    amplitude2dBSPL
    wav2dBSPL
    wav2leq
    pressure2leq
    psd2leq
    
Attenuation and active distance
-------------------------------
.. autosummary::
    :toctree: generated/
    
    attenuation_dB
    dBSPL_per_bin
    detection_distance
    pressure_at_r0
    dBSPL_at_r0
    apply_attenuation

"""

from .conversion_SPL import (wav2volt,
                        volt2pressure,
                        wav2pressure,
                        pressure2dBSPL,
                        dBSPL2pressure,
                        power2dBSPL,
                        amplitude2dBSPL,
                        wav2dBSPL,
                        wav2leq,
                        pressure2leq,
                        psd2leq)

from .active_space import (attenuation_dB,
                        dBSPL_per_bin,
                        detection_distance,
                        pressure_at_r0,
                        dBSPL_at_r0,
                        apply_attenuation)

__all__ = [
    # conversion_SPL        
    'wav2volt',
    'volt2pressure',
    'wav2pressure',
    'pressure2dBSPL',
    'dBSPL2pressure',
    'power2dBSPL',
    'amplitude2dBSPL',
    'wav2dBSPL',
    'wav2leq',
    'pressure2leq',
    'psd2leq',
    #active_space
    'attenuation_dB',
    'dBSPL_per_bin',
    'detection_distance',
    'pressure_at_r0',
    'dBSPL_at_r0',
    'apply_attenuation'
]


#!/usr/bin/env python
""" 
Mathematical tools for audio signal processing.
"""
# =============================================================================
# Load the modules
# =============================================================================
# Import external modules
import matplotlib.pyplot as plt
import numpy as np 
from numpy import mean, median, var
from scipy.ndimage import uniform_filter1d # for fast running mean
from scipy.signal import periodogram, welch
import pandas as pd
# min value
import sys
_MIN_ = sys.float_info.min

# Import internal modules
from maad.util import linear_scale


#%%
# =============================================================================
# public functions
# =============================================================================

def running_mean(x, N, mode="nearest"):
    """
    Compute fast running mean for a window size N.
    
    Parameters
    ----------
    x :  1d ndarray of scalars
        Vector 

    mode : {'reflect', 'constant', 'nearest', 'mirror', 'wrap'}, optional, 
        The `mode` parameter determines how the input array is extended
        when the filter overlaps a border. Default is 'nearest'. Behavior
        for each valid value is as follows:
    
        'reflect' (`d c b a | a b c d | d c b a`)
            The input is extended by reflecting about the edge of the last
            pixel.
    
        'constant' (`k k k k | a b c d | k k k k`)
            The input is extended by filling all values beyond the edge with
            the same constant value, defined by the `cval` parameter.
    
        'nearest' (`a a a a | a b c d | d d d d`)
            The input is extended by replicating the last pixel.
    
        'mirror' (`d c b | a b c d | c b a`)
            The input is extended by reflecting about the center of the last
            pixel.
    
        'wrap' (`a b c d | a b c d | a b c d`)
            The input is extended by wrapping around to the opposite edge.
            
    N : int
        length of window to compute the mean
        
    Returns
    -------
    x_mean : 1d ndarray of scalars
        Vector with the same dimensions than the original variable x
        
    Examples
    --------
    >>> maad.util.running_mean([2, 8, 0, 4, 1, 9, 9, 0], N=3)
        array([4, 3, 4, 1, 4, 6, 6, 3])
        
    """         
    x_mean = uniform_filter1d(x, size=N, mode="nearest")
    return x_mean

#%%
def get_unimode (X, mode ='median', axis=1, N=7, N_bins=100, verbose=False):
    """
    Get the statistical mode or modal value which is the most common number in the 
    dataset.
    
    Parameters
    ----------
    X :  1d or 2d ndarray of scalars
        Vector or matrix 
                
    mode : str, optional, default is 'median'
        Select the mode to remove the noise
        Possible values for mode are :
        - 'ale' : Adaptative Level Equalization algorithm [Lamel & al. 1981]
        - 'median' : subtract the median value
        - 'mean' : subtract the mean value (DC)
    
    axis : integer, default is 1
        if matrix, estimate the mode for each row (axis=0) or each column (axis=1)
        
    N : int (only for mode = "ale")
        length of window to compute the running mean of the histogram
        
    N_bins : int (only for mode = "ale")
        number of bins to compute the histogram
            
    verbose : boolean, optional, default is False
        print messages into the consol or terminal if verbose is True
                      
    Returns
    -------
    unimode_value : float
        The most common number in the dataset
        
    Notes
    -----
    ale : Adaptative Level Equalization algorithm from Lamel et al., 1981 :
    L.F. Lamel, L.R. Rabiner, A.E. Rosenberg, J.G. Wilpon
    An improved endpoint detector for isolated word recognition
    IEEE Trans. ASSP, ASSP-29 (1981), pp. 777-785
    `DOI: 10.1109/TASSP.1981.1163642 <https://doi.org/10.1109/TASSP.1981.1163642>`_
    
    Examples
    --------
    
    This function is interesting to obtain the background noise (BGN) profile (e.g. frequency bin
    by frequency bin) of a spectrogram
    
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> Sxx_power,tn,fn,_ = maad.sound.spectrogram(w,fs,window='hann',noverlap=512, nFFT=1024)
    >>> Sxx_dB = maad.util.power2dB(Sxx_power)
    >>> BGN_med = maad.util.get_unimode (Sxx_dB, mode='median', axis=1)
    >>> import matplotlib.pyplot as plt 
    >>> plt.plot(fn,maad.util.mean_dB(Sxx_dB,axis=1))
    >>> plt.plot(fn,BGN_med)
    
    Extract the background noise from mean
    
    >>> BGN_mean = maad.util.get_unimode (Sxx_dB, mode='mean', axis=1)
    >>> plt.plot(fn,BGN_mean)
    
    Extract the background noise from ale (i.e. unimode)
    
    >>> BGN_ale = maad.util.get_unimode (Sxx_dB, mode='ale', N=7, N_bins=100, axis=1)
    >>> plt.plot(fn,BGN_ale)
    
    """         
    if X.ndim ==2: 
        if axis == 0:
            X = X.transpose()
            axis = 1
    elif X.ndim ==1: 
        axis = 0
        
    if mode=='ale':
                
        if X.ndim ==2:
            unimode_value = []
            for i, x in enumerate(X):  
                # Min and Max of the envelope (without taking into account nan)
                x_min = np.nanmin(x)
                x_max = np.nanmax(x)
                # Compute a 50-bin histogram ranging between Min and Max values
                hist, bin_edges = np.histogram(x, bins=N_bins, range=(x_min, x_max))
                
                # smooth the histogram by running mean
                hist_smooth = running_mean(hist, N, mode="nearest")
                    
                # find the maximum of the peak with quadratic interpolation
                # don't take into account the first 4 bins.
                imax = np.argmax(hist_smooth) 
               
                unimode_value.append(bin_edges[imax])
                
            # transpose the vector
            unimode_value = np.asarray(unimode_value)
            unimode_value = unimode_value.transpose()
        else:
            x = X
            # Min and Max of the envelope (without taking into account nan)
            x_min = np.nanmin(x)
            x_max = np.nanmax(x)
            
            # Compute a 50-bin histogram ranging between Min and Max values
            hist, bin_edges = np.histogram(x, bins=N_bins, range=(x_min, x_max))
            
            # smooth the histogram by running mean
            hist_smooth = running_mean(hist, N, mode="nearest")
                                                 
            # find the maximum of the peak with quadratic interpolation
            imax = np.argmax(hist_smooth)
                        
            # assuming an additive noise model : noise_bckg is the max of the histogram
            # as it is an histogram, the value is unimode_value = bin_edges_interp[np.argmax(hist_interp)]
            unimode_value = bin_edges[imax]

    elif mode=='median':
        unimode_value = median(X, axis=axis)
        
    elif mode=='mean':
        unimode_value = mean(X, axis=axis)
        
    return unimode_value

#%%
def rms(s):
    """
    Compute the root-mean-square (RMS) level of an input signal. 
    
    RMS is defined as the square root of the arithmetic mean of the square of a set of numbers [1]. The RMS is used to estimate de mean amplitude level of an audio signal or any alternative time series.
    
    
    Parameters
    ----------
    s : 1D array
        Input signal to process

    Returns
    -------
    rms: float
        Root mean square of input signal
    
    References
    ----------
    .. [1] 'Root mean square' (2010). Wikipedia. Available at https://en.wikipedia.org/wiki/Root_mean_square
    
    Examples
    --------
    >>> from maad import sound, util
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> rms_value = util.rms(s)
    
    """
    return np.sqrt(np.mean(s**2))

#%%
def skewness (x, axis=None):
    """
    Compute the skewness (asymetry) of an audio signal.
    
    Parameters
    ----------
    x : ndarray of floats 
        1d signal or 2d matrix
    axis : integer, optional, default is None
        select the axis to compute the kurtosis
        The default is to compute the mean of the flattened array.
                            
    Returns
    -------    
    sk : float or ndarray of floats
        skewness of x 
        if x is a 1d vector => single value
        if x is a 2d matrix => array of values corresponding to the number of
        points in the other axis
    
    Examples
    --------
    >>> from maad import sound, util
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> util.skewness(s)
    -0.006547980427883208
    
    """
    if isinstance(x, (np.ndarray)) == True:
        if axis is None:
            # flatten the array
            Nf = len(np.ndarray.flatten((x)))
        else:
            Nf = x.shape[axis]
        mean_x =  np.mean(x, axis=axis)
        std_x = np.std(x, axis=axis)
        if axis == 0 :
            z = x - mean_x[np.newaxis, ...]
        else :
            z = x - mean_x[..., np.newaxis]
        sk = (np.sum(z**3, axis=axis)/(Nf-1))/std_x**3
    else:
        print ("WARNING: type of x must be ndarray") 
        sk = None

    # test if ku is an array with a single value
    if (isinstance(sk, (np.ndarray)) == True) and (len(sk) == 1):
        sk = float(sk)

    return sk

#%%
def kurtosis (x, axis=None):
    """
    Compute the kurtosis (tailedness or curved or arching) of an audio signal.
    
    Parameters
    ----------
    x : ndarray of floats 
        1d signal or 2d matrix       
    axis : integer, optional, default is None
        select the axis to compute the kurtosis
        The default is to compute the mean of the flattened array.
                            
    Returns
    -------    
    ku : float or ndarray of floats
        kurtosis of x 
        if x is a 1d vector => single value
        if x is a 2d matrix => array of values corresponding to the number of
        points in the other axis
        
    Examples
    --------
    >>> from maad import sound, util
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> util.kurtosis(s)
    24.711610834321217        
        
    """
    if isinstance(x, (np.ndarray)) == True:
        if axis is None:
            # flatten the array
            Nf = len(np.ndarray.flatten((x)))
        else:
            Nf = x.shape[axis]
        mean_x =  np.mean(x, axis=axis)
        std_x = np.std(x, axis=axis)
        if axis==0 :
            z = x - mean_x[np.newaxis, ...]
        else:
            z = x - mean_x[..., np.newaxis]
        ku = (np.sum(z**4, axis=axis)/(Nf-1))/std_x**4
    else:
        print ("WARNING: type of x must be ndarray") 
        ku = None
        
    # test if ku is an array with a single value
    if (isinstance(ku, (np.ndarray)) == True) and (len(ku) == 1):
        ku = float(ku)
       
    return ku

#%%
def moments (X, axis=None):
    """
    Computes the first 4th moments of a vector (1d, ie. spectrum or waveform) 
    or spectrogram (2d) 
    
    - mean
    - variance
    - skewness
    - kurtosis
    
    Parameters
    ----------
    X : ndarray of floats
        vector (1d : spectrum, waveform) or matrix (2d : spectrogram). 
    axis : interger, optional, default is None
        if spectrogram (2d), select the axis to estimate the moments.
        
    Returns
    -------
    mean : float 
        mean of X
    var : float 
        variance  of X
    skew : float
        skewness of X
    kurt : float
        kurtosis of X

    Examples
    --------
    >>> from maad import sound, util
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> mean, var, skew, kurt = util.moments(s)
    >>> print ('mean:%2.4f / var:%2.4f / skew:%2.4f / kurt:%2.4f' %(mean, var, skew, kurt)) 
    mean:-0.0000 / var:0.0012 / skew:-0.0065 / kurt:24.7116    
    """
    # force X to be ndarray
    X = np.asarray(X)
    
    return mean(X, axis), var(X, axis), skewness(X, axis), kurtosis(X, axis)

#%%
def entropy (x, axis=0):
    """
    Compute the entropy of a vector (waveform) or matrix (spectrogram).
    
    Parameters
    ----------
    x : ndarray of floats
        x is a vector (1d) or a matrix (2d)

    axis : int, optional, default is 0
        select the axis where the entropy is computed
        if x is a vector, axis=0
        if x is a 2d ndarray, axis=0 => rows, axis=1 => columns
                
    Returns
    -------
    H : float or ndarray of floats
        entropy of x
        
    Examples
    --------
    >>> from maad import sound, util
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> H = util.entropy(s)
    >>> print ('Entropy is %2.4f' %H) 
    Entropy is 0.9998 
        
    """
    # force x to be ndarray
    x = np.asarray(x)
    
    if isinstance(x, (np.ndarray)) == True:
        if x.ndim > axis:
            if x.shape[axis] == 0: 
                print ("WARNING: x is empty") 
                H = None 
            elif x.shape[axis] == 1:
                H = 0 # null entropy
            elif x.any() == 0: # test if there are only zeros
                if x.ndim == 1 : # case vector
                    H = 1 # entropy = 1
                else : # case matrix
                    if axis == 0 : H = np.ones(x.shape[1]) # entropy = 1
                    if axis == 1 : H = np.ones(x.shape[0]) # entropy = 1
            else:
                # if datain contains negative values -> rescale the signal between 
                # between posSitive values (for example (0,1))
                if np.min(x)<0:
                    x = linear_scale(x,minval=0,maxval=1)
                # length of datain along axis
                n = x.shape[axis]
                # Tranform the signal into a Probability mass function (pmf)
                # Sum(pmf) = 1
                if axis == 0 :
                    pmf = x/np.sum(x,axis)
                elif axis == 1 :                     
                    pmf = (x.transpose()/np.sum(x,axis)).transpose()
                pmf[pmf==0] = _MIN_
                #normalized by the length : H=>[0,1]
                H = -np.sum(pmf*np.log(pmf),axis)/np.log(n)
        else:
            print ("WARNING :axis is greater than the dimension of the array")    
            H = None 
    else:
        print ("WARNING: x must be ndarray")   
        H = None 

    return H



#!/usr/bin/env python
""" 
Ensemble of functions that facilitate data visualization.
"""

# Import external modules
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.patches as mpatches
from matplotlib.axes import Axes
import pandas as pd
from skimage.io import imsave
import colorsys
from ast import literal_eval
from datetime import datetime

# min value
import sys

_MIN_ = sys.float_info.min

# Importation from internal modules
from maad.util import linear_scale, power2dB

#%%
def _check_axes(axes):
    """Check if "axes" is an instance of an axis object. If not, use `gca`."""
    if axes is None:
        import matplotlib.pyplot as plt

        axes = plt.gca()
    elif not isinstance(axes, Axes):
        raise ValueError(
            "`axes` must be an instance of matplotlib.axes.Axes. "
            "Found type(axes)={}".format(type(axes))
        )
    return axes


#%%
def _is_single_value(value):
    """
    Check if a value is a single numeric value or a single character.

    Parameters:
    -----------
        value (int, float, str): The value to be checked.

    Returns:
    --------
        bool: True if the value is a single numeric value or a single character, False otherwise.
    """
    if isinstance(value, (int, float, str)):
        if isinstance(value, (int, float)):
            return True  # It's a numeric value
        elif isinstance(value, str):
            # Check if it's a single numeric character (0-9)
            if len(value) == 1 and value.isdigit():
                return True  # It's a single numeric character
            else:
                return False  # It's a string, but not a single character
    return False  # It's not a single value

#%%
MIN_V = 1 # min week number possible
MIN_m = 1 # min month number possible
MIN_d = 1 # min day number possible
MIN_H = 0 # min hour number possible
MIN_M = 0 # min minute number possible

MAX_V = 53 # max week number possible
MAX_m = 12 # max month number possible
MAX_d = 31 # max day number possible
MAX_H = 23 # max hour number possible
MAX_M = 59 # max minute number possible

def _filter_dataframe_by_datetime(df, date_format, date_range):
    """
    Filter a DataFrame based on a specified date format and date range.

    Parameters:
    -----------
        df : pandas.DataFrame
            The DataFrame to be filtered.
        date_format :str
            The format of the date to be extracted from the DataFrame's index.
        date_range : int, str, list
            The range of dates to filter the DataFrame by.

    Returns:
    --------
        pandas.DataFrame: The filtered DataFrame.

    Raises:
    -------
        ValueError: If the date_format is not allowed or if the date_range is not a single value or a list.

    Note:
    -----
        The date_format parameter can be one of the following:
        - "%V": Week number
        - "%Y": Year
        - "%m-%d": Month and day
        - "%a": Weekday
        - "%m": Month
        - "%d": Day
        - "%y-%m-%d": Year, Month, and Day
        - "%H:%M": Hour and minute (e.g., "13:42")

        The date_range parameter can be one of the following:
        - A single value (int or str) representing a specific date or time.
        - A list of values (int or str) representing a range of dates or times.
        For example, ["01-01", "01-31"] represents a range of dates from January 1st to January 31st.

    """
    # Extract the date portion from the DatetimeIndex based on the specified date format
    try:
        df['date_type'] = df.index.strftime(date_format)
    except :
        raise ValueError('Error in the date_format. Check the date_format parameter')

    if _is_single_value(date_range) == True :
        if (date_format == "%V") or (date_format == "%Y") or (date_format == "%m") or (date_format == "%d"):
            # Filter by week number or by year or by month
            df_filtered = df[df['date_type'].astype(int) == date_range]
        elif (date_format == "%m-%d") or (date_format == "%H:%M") or (date_format == "%a") or (date_format == "%y-%m-%d"):
            # Filter by time (hour:minutes)
            # Filter by weekday (e.g., "Mon" for Monday)
            # Filter by year-month-day (e.g., "23-01-01" for January 1, 2023)
            # Filter by month and day (e.g., "01-01" for January 1)
            df_filtered = df[df['date_type'] == date_range]
        else :
            print("Filtering was no possible. date_format {} is not allowed".format(date_format))
            df_filtered = df

    elif (len(date_range)==1) or(len(date_range)>2):
        if (date_format == "%V") or (date_format == "%Y") or (date_format == "%m") or (date_format == "%d"):
            # Filter by week number or by year or by month
            df_filtered = df[df['date_type'].astype(int).isin(date_range)]
        elif (date_format == "%m-%d") or (date_format == "%H:%M") or (date_format == "%a") or (date_format == "%y-%m-%d"):
            # Filter by time (hour:minutes)
            # Filter by weekday (e.g., "Mon" for Monday)
            # Filter by year-month-day (e.g., "23-01-01" for January 1, 2023)
            # Filter by month and day (e.g., "01-01" for January 1)
            df_filtered = df[df['date_type'].isin(date_range)]
        else :
            print("Filtering was no possible. date_format {} is not allowed".format(date_format))
            df_filtered = df
    else:
        if isinstance(date_range, list) : 
            if (date_format == "%V") or (date_format == "%Y") or (date_format == "%m") or (date_format == "%d"):
                # Filter by week number or by year or by month
                df_filtered = df[df['date_type'].astype(int).between(date_range[0], date_range[1])]
            elif (date_format == "%m-%d") or (date_format == "%H:%M") or (date_format == "%a") or (date_format == "%y-%m-%d"):
                # Filter by time (hour:minutes)
                # Filter by weekday (e.g., "Mon" for Monday)
                # Filter by year-month-day (e.g., "23-01-01" for January 1, 2023)
                # Filter by month and day (e.g., "01-01" for January 1)
                if date_range[0] <= date_range[1] :
                    df_filtered = df[df['date_type'].between(str(date_range[0]), str(date_range[1]))]
                else :
                    if date_format == "%H:%M" : 
                        MAX = f'{MAX_H:02}:{MAX_M:02}'
                        df_filtered_part1 = df[df['date_type'].between(str(date_range[0]), MAX)] 
                        MIN = f'{MIN_H:02}:{MIN_M:02}'
                        df_filtered_part2 = df[df['date_type'].between(MIN, str(date_range[1]))]  
                        df_filtered = pd.concat([df_filtered_part1, df_filtered_part2], axis = 0)
                    else :
                        print("Filtering was no possible. The range must increase")
                        df_filtered = df
            else :
                print("Filtering was no possible. date_format {} is not allowed".format(date_format))
                df_filtered = df
        else :
            print("Filtering was no possible. date_range must be a single value or a list")
            df_filtered = df

    # Check if 'data_type' is in the DataFrame's columns
    if 'date_type' in df_filtered.columns:
        # If it exists, drop the column
        df_filtered = df_filtered.drop('date_type', axis=1)

    return df_filtered

#%%
def plot_shape(shape, params, row=0, ax=None, display_values=False):
    """ 
    Plot shape features in a bidimensional plot.
    
    Parameters 
    ---------- 
    shape: 1D array, pd.Series or pd.DataFrame 
        Shape features computed with shape_features function. 
    
    params: pd.DataFrame 
        Pandas dataframe returned by maad.features_rois.shape_features 
    
    row: int 
        Observation to be visualized 
    
    display_values: bool 
        Set to True to display the coefficient values. Default is False. 
    
    Returns 
    ------- 
    ax: matplotlib.axes 
        Axes of the figure 
    
    Examples 
    -------- 
    >>> from maad.sound import load, spectrogram 
    >>> from maad.features import shape_features, plot_shape 
    >>> import numpy as np 
    >>> s, fs = load('../data/spinetail.wav') 
    >>> Sxx, ts, f, ext = spectrogram(s, fs) 
    >>> shape, params = shape_features(np.log10(Sxx), resolution='high') 
    >>> plot_shape(shape, params) 

    """

    # compute shape of matrix
    dirs_size = params.theta.unique().shape[0]
    scale_size = np.unique(params.freq).size * np.unique(params.pyr_level).size
    # reshape feature vector
    idx = params.sort_values(["theta", "pyr_level", "scale"]).index

    if isinstance(shape, pd.DataFrame):
        shape_plt = shape.iloc[:, shape.columns.str.startswith("shp")]
        shape_plt = np.reshape(shape_plt.iloc[row, idx].values, (dirs_size, scale_size))
    elif isinstance(shape, pd.Series):
        shape_plt = shape.iloc[shape.index.str.startswith("shp")]
        shape_plt = np.reshape(shape_plt.iloc[idx].values, (dirs_size, scale_size))
    elif isinstance(shape, np.ndarray):
        shape_plt = np.reshape(shape_plt[idx], (dirs_size, scale_size))

    unique_scale = params.scale[idx] * 2 ** params.pyr_level[idx]
    # get textlab
    textlab = shape_plt
    textlab = np.round(textlab, 2)

    # plot figure
    if ax is None:
        fig = plt.figure(figsize=(8, 6))
        ax = fig.add_subplot(111)
    else:
        pass
    ax.imshow(
        shape_plt, aspect="auto", origin="lower", interpolation="None", cmap="viridis"
    )
    if display_values:
        for (j, i), label in np.ndenumerate(textlab):
            ax.text(i, j, label, ha="center", va="center")

    yticklab = np.round(params.theta.unique(), 1)
    xticklab = np.reshape(unique_scale.values, (dirs_size, scale_size))
    ax.set_xticks(np.arange(scale_size))
    ax.set_xticklabels(np.round(xticklab, 1)[0, :])
    ax.set_yticks(np.arange(dirs_size))
    ax.set_yticklabels(yticklab)
    ax.set_xlabel("Scale")
    ax.set_ylabel("Theta")
    plt.show()

    return ax

#%%
def overlay_centroid(im_ref, centroid, savefig=None, **kwargs):
    """ 
    Overlay centroids on the original spectrogram 
    
    Parameters 
    ---------- 
    Sxx :  2D array 
        Spectrogram 
        
    centroid: pandas DataFrame 
        DataFrame with centroid descriptors (centroid_f, centroid_t) 
        Do format_features(rois,tn,fn) before using overlay_centroid to be sure that 
        the format of the DataFrame is correct 
            
    savefig : string, optional, default is None 
        Root filename (with full path) is required to save the figures. Postfix 
        is added to the root filename. 
        
    kwargs, optional. This parameter is used by plt.plot and savefig functions 
            
        - savefilename : str, optional, default :'_spectro_overlaycentroid.png' 
            Postfix of the figure filename 
        
        - extent : list of scalars [left, right, bottom, top], optional, default: None
            The location, in data-coordinates, of the lower-left and
            upper-right corners. If `None`, the image is positioned such that
            the pixel centers fall on zero-based (row, column) indices.
        
        - title : string, optional, default : 'Spectrogram' 
            title of the figure 
        
        - xlabel : string, optional, default : 'Time [s]' 
            label of the horizontal axis 
        
        - ylabel : string, optional, default : 'Amplitude [AU]' 
            label of the vertical axis 
        
        - cmap : string or Colormap object, optional, default is 'gray' 
            See https://matplotlib.org/examples/color/colormaps_reference.html 
            in order to get all the  existing colormaps 
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic',  
            'viridis'... 
        
        - vmin, vmax : scalar, optional, default: None 
            `vmin` and `vmax` are used in conjunction with norm to normalize 
            luminance data.  Note if you pass a `norm` instance, your 
            settings for `vmin` and `vmax` will be ignored. 
        
        - dpi : integer, optional, default is 96 
            Dot per inch.  
            For printed version, choose high dpi (i.e. dpi=300) => slow 
            For screen version, choose low dpi (i.e. dpi=96) => fast 
        
        - format : string, optional, default is 'png' 
            Format to save the figure  
        
        ... and more, see matplotlib  

    Returns 
    ------- 
    ax  
        axis object (see matplotlib) 
    fig  
        figure object (see matplotlib) 

    Examples
    --------

    Get centroid from the whole power spectrogram 

    >>> from maad.sound import load, spectrogram
    >>> from maad.features import centroid_features
    >>> from maad.util import (power2dB, format_features, overlay_rois, plot2d,
                            overlay_centroid)
    
    Load audio and compute spectrogram 
    
    >>> s, fs = load('../data/spinetail.wav') 
    >>> Sxx,tn,fn,ext = spectrogram(s, fs, db_range=80) 
    >>> Sxx = power2dB(Sxx, db_range=80)
    
    Load annotations and plot
    
    >>> from maad.util import read_audacity_annot
    >>> rois = read_audacity_annot('../data/spinetail.txt') 
    >>> rois = format_features(rois, tn, fn) 
    >>> ax, fig = plot2d (Sxx, extent=ext)
    >>> ax, fig = overlay_rois(Sxx,rois, extent=ext, ax=ax, fig=fig)
    
    Compute the centroid of each rois, format to get results in the 
    temporal and spectral domain and overlay the centroids.
    
    >>> centroid = centroid_features(Sxx, rois) 
    >>> centroid = format_features(centroid, tn, fn)
    >>> ax, fig = overlay_centroid(Sxx,centroid, extent=ext, ax=ax, fig=fig)
    
    """
    
    # Check format of the input data
    if type(centroid) is not pd.core.frame.DataFrame:
        raise TypeError("Rois must be of type pandas DataFrame")

    if not (("centroid_t" and "centroid_f") in centroid):
        raise TypeError(
            "Array must be a Pandas DataFrame with column names:(centroid_t, centroid_f). Check example in documentation."
        )

    ylabel = kwargs.pop("ylabel", "Frequency [Hz]")
    xlabel = kwargs.pop("xlabel", "Time [s]")
    title = kwargs.pop("title", "ROIs Overlay")
    cmap = kwargs.pop("cmap", "gray")
    extent = kwargs.pop("ext", None)

    if extent is None:
        ylabel = "pseudofrequency [points]"
        xlabel = "pseudotime [points]"

    vmin = kwargs.pop("vmin", 0)
    vmax = kwargs.pop("vmax", 1)
    ax = kwargs.pop("ax", None)
    fig = kwargs.pop("fig", None)
    color = kwargs.pop("color", "firebrick")
    ms = kwargs.pop("ms", 2)
    marker = kwargs.pop("marker", "o")

    if (ax is None) and (fig is None):
        ax, fig = plot2d(
                        im_ref,
                        extent  = extent,
                        now     = False,
                        title   = title,
                        ylabel  = ylabel,
                        xlabel  = xlabel,
                        vmin    = vmin,
                        vmax    = vmax,
                        cmap    = cmap,
                        **kwargs
                        )

    ax.plot(centroid.centroid_t, centroid.centroid_f, marker, ms=ms, color=color)

    fig.canvas.draw()

    # SAVE FIGURE
    if savefig is not None:
        dpi = kwargs.pop("dpi", 96)
        format = kwargs.pop("format", "png")
        filename = kwargs.pop("filename", "_spectro_overlaycentroid")
        filename = savefig + filename + "." + format
        fig.savefig(filename, bbox_inches="tight", dpi=dpi, format=format, **kwargs)

    return ax, fig

#%%
def overlay_rois(im_ref, rois, edge_color=None, unique_labels= None, 
                 textbox_label=False, savefig=None, **kwargs):
    """ 
    Display bounding boxes with time-frequency regions of interest over a spectrogram.
    
    Regions of interest (ROIs) must be provided. They can be loaded as manual annotations or computed using automated methods (see the example section).
    
    Parameters 
    ---------- 
    im_ref : 2d ndarray of scalars 
        Spectrogram (or image) 

    rois_bbox : pandas.DataFrame
        Contains the bounding box of each ROI. The pandas.DataFrame must have columns
        ['min_x', 'max_x', 'min_y', 'max_y']. If your data is in the time-frequency
        format ['min_t', 'max_t', 'min_f', 'max_f'], use the function
        maad.util.format_features to get the cooresponding
        x, y coordinates.
        
    edge_color : {None, string, tuple, list of tuples}, default is None
        Define the color of the bounding-box to display.
        if it's a string, select a color name 
        if it's a tuple, give a tuple of floats [0,1] in the form (r,g,b,a), 
        with r for red, g for green, b for blue and a for alpha (i.e transparency)
        for example
        - pure red : (1,0,0,0)
        - pure yellow : (1,1,0,0)
        - pure blue with transparency(0,0,1,0.5)
        if it's a list, give a list of tuple or string. The number of element 
        in the list should be equal to the number of elements in unique_labels.
        if it's set to None, a random list of colors is used.
        
    unique_labels : list, default is None
        list of unique labels that could be either strings or numbers. When
        a list is given as argument, the number element in the list should 
        be the same as the number of edge_color
    
    textbox_label: bool
        Display a text box above the bounding box indicating the name of the label.
        The rois_bbox pandas.DataFrame must have a column called 'label' with names
        assigned to each ROI.

    savefig : string, optional, default is None 
        Root filename (with full path) is required to save the figures. Postfix 
        is added to the root filename. 
        
    kwargs, optional. This parameter is used by plt.plot and savefig functions 
            
        - savefilename : str, optional, default :'_spectro_overlayrois.png' 
            Postfix of the figure filename 
            
        - extent : list of scalars [left, right, bottom, top], optional, default: None
            The location, in data-coordinates, of the lower-left and
            upper-right corners. If `None`, the image is positioned such that
            the pixel centers fall on zero-based (row, column) indices.  
        
        - title : string, optional, default : 'Spectrogram' 
            title of the figure 
        
        - xlabel : string, optional, default : 'Time [s]' 
            label of the horizontal axis 
        
        - ylabel : string, optional, default : 'Amplitude [AU]' 
            label of the vertical axis 
        
        - cmap : string or Colormap object, optional, default is 'gray' 
            See https://matplotlib.org/examples/color/colormaps_reference.html 
            in order to get all the  existing colormaps 
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic',  
            'viridis'... 
        
        - vmin, vmax : scalar, optional, default: None 
            `vmin` and `vmax` are used in conjunction with norm to normalize 
            luminance data.  Note if you pass a `norm` instance, your 
            settings for `vmin` and `vmax` will be ignored. 
        
        - dpi : integer, optional, default is 96 
            Dot per inch.  
            For printed version, choose high dpi (i.e. dpi=300) => slow 
            For screen version, choose low dpi (i.e. dpi=96) => fast 
        
        - format : string, optional, default is 'png' 
            Format to save the figure  
        
        ... and more, see matplotlib  

    Returns 
    ------- 
    ax : axis object (see matplotlib) 
        
    fig : figure object (see matplotlib) 
    
    Examples 
    -------- 
    
    Simple display of bounding boxes over the spectrogram.
    
    >>> from maad import sound, util
    >>> s, fs = sound.load('./data/spinetail.wav')
    >>> rois = util.read_audacity_annot('./data/spinetail.txt')
    >>> Sxx, tn, fn, ext = sound.spectrogram(s, fs, nperseg=512, noverlap=256)
    >>> rois = util.format_features(rois, tn, fn)
    >>> fig, ax = plt.subplots(figsize=(10,5))
    >>> util.plot_spectrogram(Sxx,ext, db_range=70, ax=ax)
    >>> util.overlay_rois(Sxx, rois, fig=fig, ax=ax, textbox_label=True)
    
    Detect regions of interest and display them over the spectrogram.
    Load audio recording and compute the spectrogram.
    
    >>> s, fs = maad.sound.load('../data/cold_forest_daylight.wav')
    >>> Sxx,tn,fn,ext = maad.sound.spectrogram (s, fs, fcrop=(0,10000))   

    Subtract the background noise before finding ROIs.
    
    >>> Sxx_noNoise = maad.sound.median_equalizer(Sxx)
    
    Convert linear spectrogram into dB and smooth.
    
    >>> Sxx_noNoise_dB = maad.util.power2dB(Sxx_noNoise)         
    >>> Sxx_noNoise_dB_blurred = maad.sound.smooth(Sxx_noNoise_dB)
    
    Detection of the acoustic signature => creation of a mask
    
    >>> im_bin = maad.rois.create_mask(Sxx_noNoise_dB_blurred, bin_std=6, bin_per=0.5, mode='relative') 
    
    Select rois from the mask and display bounding box over the spectrogram without noise.
    
    >>> import numpy as np
    >>> im_rois, df_rois = maad.rois.select_rois(im_bin, min_roi=100)  
    >>> maad.util.overlay_rois (Sxx_noNoise_dB, df_rois, extent=ext,vmin=np.median(Sxx_noNoise_dB), vmax=np.median(Sxx_noNoise_dB)+60) 
        
    """

    # Check format of the input data
    if type(rois) is not pd.core.frame.DataFrame:
        raise TypeError("Rois must be of type pandas DataFrame.")
    if not (("min_y" and "min_x" and "max_y" and "max_x") in rois):
        raise TypeError(
            "Array must be a Pandas DataFrame with column names: min_y, min_x, max_y, max_x. Check example in documentation."
        )
        
    ylabel = kwargs.pop("ylabel", "Frequency [Hz]")
    xlabel = kwargs.pop("xlabel", "Time [s]")
    title = kwargs.pop("title", "ROIs Overlay")
    cmap = kwargs.pop("cmap", "gray")
    vmin = kwargs.pop("vmin", np.percentile(im_ref, 0.05))
    vmax = kwargs.pop("vmax", np.percentile(im_ref, 0.95))
    ax = kwargs.pop("ax", None)
    fig = kwargs.pop("fig", None)
    extent = kwargs.pop("extent", None)

    if extent is None:
        xlabel = "pseudoTime [points]"
        ylabel = "pseudoFrequency [points]"
        
    if (ax is None) and (fig is None):
        ax, fig = plot2d(
                        im_ref,
                        extent  = extent,
                        now     = False,
                        title   = title,
                        ylabel  = ylabel,
                        xlabel  = xlabel,
                        vmin    = vmin,
                        vmax    = vmax,
                        cmap    = cmap,
                        **kwargs
                        )

    # Convert pixels into time and frequency values
    y_len, x_len = im_ref.shape
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    x_scaling = (xmax - xmin) / x_len
    y_scaling = (ymax - ymin) / y_len

    # if no list of unique labels is provided, try to find a "label" column
    if unique_labels is None :
        # test if rois has a label column
        if "label" in rois:
            # select the label column
            rois_label = rois.label.values
            uniqueLabels = list(np.unique(np.array(rois_label)))
        else:
            uniqueLabels = [0]    
    else: 
        if isinstance(unique_labels,list) :     
            uniqueLabels = unique_labels
        elif isinstance(unique_labels,np.ndarray) :     
            uniqueLabels = list(unique_labels)
        elif isinstance(unique_labels, str) :
            uniqueLabels = [unique_labels]
        else:
            # test if unique_labels is a number
            try:
                int(unique_labels)
                uniqueLabels = [unique_labels]
            except:
                raise TypeError(
                    "unique_labels must be a list or a string or a number"
                )      
        
    # Colormap
    if edge_color is None :
        # in order to still have the color yellow by default in case of no label
        # or a unique label
        if len(uniqueLabels) == 1:
            color = tuple(["yellow"])
        else:
            color_func = rand_cmap(len(uniqueLabels) + 1, first_color_black=False)
            color = color_func(np.arange(len(uniqueLabels)))
    # if a list of colors is defined, test if the length of the list is 
    # enough compared to the number of unique labels
    # convert into tuple
    elif isinstance(edge_color, str) :    
        color = tuple([edge_color]) * len(uniqueLabels)
    elif isinstance(edge_color, list) :  
        nn = int(len(uniqueLabels)/len(edge_color)) +1
        color = tuple(edge_color) * nn

    cc = 0
    for index, row in rois.iterrows():
        cc = cc + 1
        y0 = row["min_y"]
        x0 = row["min_x"]
        y1 = row["max_y"]
        x1 = row["max_x"]
        # find the position of the label in uniqueLabels
        if "label" in row :
            ii = uniqueLabels.index(row["label"])
        else:
            ii = 0
        rect = mpatches.Rectangle(
            (x0 * x_scaling + xmin, y0 * y_scaling + ymin),
            (x1 - x0) * x_scaling,
            (y1 - y0) * y_scaling,
            fill=False,
            edgecolor=color[ii],
            linewidth=1,
        )
        # draw the rectangle
        ax.add_patch(rect)
        
        if textbox_label:
            textbox = dict(boxstyle='square', fc=color[ii], 
                        ec=color[ii], alpha=1, pad=0, linewidth=2)
            ax.text(x0*x_scaling + xmin, (y1+2)*y_scaling + ymin, 
                    str(row.label), 
                    color='white',
                    fontweight='semibold',
                    fontsize='small',
                    fontfamily='sans-serif',
                    fontvariant='small-caps',
                    bbox=textbox
                    )
            
    if fig is not None:
        fig.canvas.draw()

    # SAVE FIGURE
    if savefig is not None and fig is not None:
        dpi = kwargs.pop("dpi", 96)
        format = kwargs.pop("format", "png")
        filename = kwargs.pop("filename", "_spectro_overlayrois")
        filename = savefig + filename + "." + format
        fig.savefig(filename, bbox_inches="tight", dpi=dpi, format=format, **kwargs)

    return ax, fig

#%%
def plot1d(x, y, ax=None, **kwargs):
    """
    Plot the waveform or spectrum of an audio signal. 
    
    Parameters
    ----------
    x : 1d ndarray of integer
        Vector containing the abscissa values (horizontal axis)
        
    y : 1d ndarray of scalar
        Vector containing the ordinate values (vertical axis)  
    
    ax : axis, optional, default is None
        Draw the signal on this specific axis. Allow multiple plots on the same
        axis.
            
    kwargs, optional
        - figsize : tuple of integers, optional, default: (4,10)
            width, height in inches.  
        - facecolor : matplotlib color, optional, default: 'w' (white)
            the background color.  
        - edgecolor : matplotlib color, optional, default: 'k' (black)
            the border color. 
        - color : matplotlib color, optional, default: 'k' (black)
            the line color
            The following color abbreviations are supported:
            ==========  ========
            character   color
            ==========  ========
            'b'         blue
            'g'         green
            'r'         red
            'c'         cyan
            'm'         magenta
            'y'         yellow
            'k'         black
            'w'         white
            ==========  ========
            In addition, you can specify colors in many ways, including RGB 
            tuples (0.2,1,0.5). See matplotlib color
        - linewidth : scalar, optional, default: 0.5
            width in pixels
        - figtitle: string, optional, default: 'Audiogram'
            Title of the plot 
        - xlabel : string, optional, default : 'Time [s]'
            label of the horizontal axis
        - ylabel : string, optional, default : 'Amplitude [AU]'
            label of the vertical axis
        - legend : string, optional, default : None
            Legend for the plot
        - now : boolean, optional, default : True
            if True, display now. Cannot display multiple plots. 
            To display mutliple plots, set now=False until the last call for 
            the last plot         
        
        ...and more, see matplotlib
        
    Returns
    -------
    fig : Figure
        The Figure instance 
    ax : Axis
        The Axis instance
        
    Examples
    --------
    >>> import numpy as np    
    >>> s, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    
    Plot the audiogram
    
    >>> tn = np.arange(0,len(s))/fs
    >>> fig, ax = maad.util.plot1d(tn,s)
    
    >>> Sxx_power,tn,fn,_ = maad.sound.spectrogram(s,fs)
    
    Convert spectrogram into dB SPL
    
    >>> Lxx = maad.spl.power2dBSPL(Sxx_power, gain=42) 
    
    Plot the spectrum at t = 7s
    
    >>> index = maad.util.nearest_idx(tn,7)
    >>> fig_kwargs = {'figtitle':'Spectrum (PSD)',
                    'xlabel':'Frequency [Hz]',
                    'ylabel':'Power [dB]',
                    'linewidth': 0.5
                    }
    
    >>> fig, ax = maad.util.plot1d(fn, Lxx[:,index], **fig_kwargs)
    """

    figsize = kwargs.pop("figsize", (4, 10))
    facecolor = kwargs.pop("facecolor", "w")
    edgecolor = kwargs.pop("edgecolor", "k")
    linewidth = kwargs.pop("linewidth", 0.5)
    color = kwargs.pop("color", "k")
    title = kwargs.pop("figtitle", "")
    xlabel = kwargs.pop("xlabel", "Time [s]")
    ylabel = kwargs.pop("ylabel", "Amplitude [AU]")
    legend = kwargs.pop("legend", None)
    now = kwargs.pop("now", False)

    # if no ax, create a figure and a subplot associated a figure otherwise
    # find the figure that belongs to ax
    if ax is None:
        fig, ax = plt.subplots()
        # set the paramters of the figure
        fig.set_figheight(figsize[0])
        fig.set_figwidth(figsize[1])
        fig.set_facecolor(facecolor)
        fig.set_edgecolor(edgecolor)
    else:
        fig = ax.get_figure()

    # plot the data on the subplot
    line = ax.plot(x, y, color=color, linewidth=linewidth, **kwargs)

    # set legend to the line
    line[0].set_label(legend)

    # set the parameters of the subplot
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    #ax.axis("tight")
    ax.grid(True, linewidth=0.3)
    ax.margins(x=0)
    
    if legend is not None:
        ax.legend()

    # Display the figure now
    if now:
        plt.show()

    return ax, fig

#%%
def plot_wave(s, fs, tlims=None, ax=None, **kwargs):
    """
    Plot audio waveform.
    
    Parameters
    ----------
    s : 1d ndarray
        Audio signal.
    fs : int
        Sampling rate of audio signal.
    tlims : tuple, optional
        Minimum and maximum temporal limits for the display (min_t, max_t). 
        The default is None.
    ax : matplotlib.axes, optional
        Pre-existing axes for the plot. The default is None.
    **kwargs : matplotlib figure properties
        Other keyword arguments that are passed down to matplotlib.axes.

    Returns
    -------
    ax : matplotlib.axes
        The matplotlib axes associated to plot.
    
    See also
    --------
    maad.sound.plot_spectrum
        
    Examples
    --------
    
    Plot a waveform of an audio signal.
    >>> from maad import sound
    >>> from maad.util import plot_wave
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> ax = plot_wave(s, fs)
    
    Use `plot_wave` with predifined matplotlib axes.
    >>> import matplotlib.pyplot as plt
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> fig, ax = plt.subplots(2,1)
    >>> plot_wave(s, fs, ax=ax[0], xlabel='', figtitle='Spinetail')
    >>> plot_wave(s, fs, tlims=(5,8), ax=ax[1])
    """

    if tlims is not None:
        s = s[int(tlims[0] * fs) : int(tlims[1] * fs)]
        t = np.linspace(0, s.shape[0] / fs, s.shape[0])
        t = t + tlims[0]  # add minimum value
    else:
        t = np.linspace(0, s.shape[0] / fs, s.shape[0])

    ax, fig = plot1d(t, s, ax, **kwargs)

    return ax

#%%
def plot_spectrum(pxx, f_idx, ax=None, flims=None, log_scale=False, fill=True, **kwargs):
    """
    Plot power spectral density estimate (PSD).
    
    Parameters
    ----------
    pxx : 1d ndarray
        Power spectral density estimate computed with `maad.sound.spectrum`.
        
    f_idx : 1d ndarray
        Index of frequencies associated with the PSD.
    
    ax : matplotlib.axes, optional
        Pre-existing axes for the plot. The default is None.
        
    flims : tuple, optional
        Minimum and maximum spectral limits for the display (min_f, max_f). 
        Default is None.

    log_scale : bool, optional
        Use a logarithmic scale to display amplitude values. The default is False.
    
    fill : bool, optional
        Fill the area between the curve and the minimum value.
        
    **kwargs : matplotlib figure properties
            Other keyword arguments that are passed down to matplotlib.axes.

    Returns
    -------
    ax : matplotlib.axes
        The matplotlib axes associated to plot.

    See also
    --------
    maad.sound.spectrum, maad.util.plot_wave
    
    Examples
    --------
    
    Plot a spectrum of an audio signal.
    
    >>> from maad import sound, util
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> pxx, f_idx = sound.spectrum(s, fs, nperseg=1024)
    >>> util.plot_spectrum(pxx, f_idx)
    
    Use `plot_spectrum` with predifined matplotlib axes.
    
    >>> import matplotlib.pyplot as plt
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> s_slice = sound.trim(s, fs, 5, 8)
    >>> pxx, f_idx = sound.spectrum(s_slice, fs, nperseg=1024)
    >>> fig, ax = plt.subplots(2,1, figsize=(10,6))
    >>> util.plot_wave(s_slice, fs, ax=ax[0])
    >>> util.plot_spectrum(pxx, f_idx, ax=ax[1], log_scale=True)
    
    """
    ylabel = kwargs.pop("ylabel", "Amplitude [AU]")
    xlabel = kwargs.pop("xlabel", "Frequency [Hz]")

    if log_scale == True:
        pxx = power2dB(pxx)
        ylabel = 'Amplitude [dB]'

    if flims is not None:
        pxx = pxx[(f_idx >= flims[0]) & (f_idx <= flims[1])]
        f_idx = f_idx[(f_idx >= flims[0]) & (f_idx <= flims[1])]

    amp_min = pxx.min()
    ax, fig = plot1d(f_idx, pxx, ax=ax, ylabel=ylabel, xlabel=xlabel, **kwargs)
    if fill:
        ax.fill_between(f_idx, amp_min, pxx, alpha=0.3, fc="grey")

    return ax

#%%
def plot2d(im, ax=None, colorbar=True, **kwargs):
    """
    Display the spectrogram of an audio signal. 
    
    The spectrogram should be previously computed using the function 
    ``maad.sound.spectrogram``.
    
    Parameters
    ----------
    im : 2D numpy array 
        Image or Spectrogram
        
    ax : axis, optional, default is None
        Draw the image on this specific axis. Allow multiple plots on the same
        figure.
            
    kwargs, optional
        - figsize : tuple of integers, optional, default: (4,13)
            width, height in inches.  
        - title : string, optional, default : 'Spectrogram'
            title of the figure
        - xlabel : string, optional, default : 'Time [s]'
            label of the horizontal axis
        - ylabel : string, optional, default : 'Frequency [Hz]'
            label of the vertical axis
        - xticks : tuple of ndarrays, optional, default : none
            * ticks : array_like => A list of positions at which ticks should be placed. You can pass an empty list to disable yticks.
            * labels : array_like, optional =>  A list of explicit labels to place at the given locs.
        - yticks : tuple of ndarrays, optional, default : none
            * ticks : array_like => A list of positions at which ticks should be placed. You can pass an empty list to disable yticks.
            * labels : array_like, optional =>  A list of explicit labels to place at the given locs.
        - cmap : string or Colormap object, optional, default is 'gray'
            See https://matplotlib.org/examples/color/colormaps_reference.html
            in order to get all the  existing colormaps
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic', 
            'viridis'...
        - vmin, vmax : scalar, optional, default: None
            `vmin` and `vmax` are used in conjunction with norm to normalize
            luminance data.  Note if you pass a `norm` instance, your
            settings for `vmin` and `vmax` will be ignored.
        - extent : list of scalars [left, right, bottom, top], optional, default: None
            The location, in data-coordinates, of the lower-left and
            upper-right corners. If `None`, the image is positioned such that
            the pixel centers fall on zero-based (row, column) indices.
        - now : boolean, optional, default : True
            if True, display now. Cannot display multiple images. 
            To display mutliple images, set now=False until the last call for 
            the last image    
        - interpolation : list of string [None, 'none', 'nearest', 'bilinear', 
            'bicubic', 'spline16','spline36', 'hanning', 'hamming', 'hermite', 
            'kaiser', 'quadric','catrom', 'gaussian', 'bessel', 'mitchell', 
            'sinc', 'lanczos'], optional, default : None
            
        ... and more, see matplotlib
        
    Returns
    -------
    fig : Figure
        The Figure instance 
    ax : Axis
        The Axis instance
        
    Examples
    --------
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> Sxx_power,tn,fn,_ = maad.sound.spectrogram(w,fs)
    >>> Lxx = maad.spl.power2dBSPL(Sxx_power, gain=42) # convert into dB SPL
    >>> fig_kwargs = {'vmax': Lxx.max(),
                    'vmin':0,
                    'extent':(tn[0], tn[-1], fn[0], fn[-1]),
                    'title':'Power spectrogram density (PSD) in dB SPL',
                    'xlabel':'Time [s]',
                    'ylabel':'Frequency [Hz]',
                    }
    >>> fig, ax = maad.util.plot2d(Lxx,interpolation=None,**fig_kwargs)      
        
    """

    # matplotlib parameters
    title = kwargs.pop("title", "")
    ylabel = kwargs.pop("ylabel", "Frequency [Hz]")
    xlabel = kwargs.pop("xlabel", "Time [s]")
    xticks = kwargs.pop("xticks", None)
    yticks = kwargs.pop("yticks", None)
    cmap = kwargs.pop("cmap", "gray")
    vmin = kwargs.pop("vmin", None)
    vmax = kwargs.pop("vmax", None)
    extent = kwargs.pop("extent", None)
    now = kwargs.pop("now", False)

    if extent is not None:
        figsize = kwargs.pop(
            "figsize",
            (0.20 * (extent[3] - extent[2]) / 1000, 0.33 * (extent[1] - extent[0])),
        )
    else:
        figsize = kwargs.pop("figsize", (4, 13))

    # if no ax, create a figure and a subplot associated a figure otherwise
    # find the figure that belongs to ax
    if ax is None:
        fig, ax = plt.subplots()
        # set the paramters of the figure
        fig.set_facecolor("w")
        fig.set_edgecolor("k")
        fig.set_figheight(figsize[0] + 1)
        fig.set_figwidth(figsize[1])
    else:
        fig = ax.get_figure()

    # display image
    _im = ax.imshow(im, extent=extent, origin="lower", cmap=cmap, 
                    vmin=vmin, vmax=vmax, **kwargs)
    if colorbar==True:
        plt.colorbar(_im, ax=ax)

    # set the parameters of the subplot
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if xticks is not None:
        ax.set_xticks(ticks=xticks[0])
        if len(xticks) == 2:
            ax.set_xticklabels(labels=xticks[1])
    if yticks is not None:
        ax.set_yticks(ticks=yticks[0])
        if len(yticks) == 2:
            ax.set_yticklabels(labels=yticks[1])
    ax.axis("tight")

    fig.tight_layout()

    # Display the figure now
    if now:
        plt.show()

    return ax, fig

#%%
def plot_spectrogram(Sxx, extent, db_range=96, gain=0, log_scale=True,
                     colorbar=True, interpolation='bilinear', **kwargs):
    """
    Plot spectrogram represenation.
    
    Parameters
    ----------
    Sxx : 2d ndarray
        Spectrogram computed using `maad.sound.spectrogram`.
    extent : list of scalars [left, right, bottom, top]
        Location, in data-coordinates, of the lower-left and upper-right corners.
    db_range : int or float, optional
        Range of values to display the spectrogram. The default is 96.
    gain : int or float, optional
        Gain in decibels added to the signal for display. The default is 0.
    log_scale : bool, optional
        Logarithmic transformation of spectrogram coefficients to enhace visual 
        representation. The default is True.
    colorbar : bool, optional
        Plot the colorbar next to the spectrogram. The default is True.
    interpolation : bool, optional
        Interpolate spectrogram values to enhance visual representation. 
        The default is 'bilinear'.
    **kwargs : matplotlib figure properties
            Other keyword arguments that are passed down to matplotlib.axes.

    Returns
    -------
    ax : matplotlib.axes
        The matplotlib axes associated to plot.

    See also
    --------
    maad.sound.plot_spectrum, maad.util.plot_wave
    
    Examples
    --------
    
    Plot the spectrogram of an audio.
    
    >>> from maad import sound, util
    >>> s, fs = sound.load('../data/spinetail.wav') 
    >>> Sxx, tn, fn, ext = sound.spectrogram(s,fs)
    >>> util.plot_spectrogram(Sxx, ext, db_range=50, gain=30, figsize=(4,10))
        
    Use `plot_spectrogram` with predifined matplotlib axes.
    
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots(2,1, figsize=(10,6))
    >>> util.plot_wave(s, fs, ax=ax[0])
    >>> util.plot_spectrogram(Sxx, ext, db_range=50, gain=30, colorbar=False, ax=ax[1])

    Plot a single frase of the spinetail.
    
    >>> Sxx, tn, fn, ext = sound.spectrogram(s,fs, flims=(2000,15000), tlims=(5,8))
    >>> util.plot_spectrogram(Sxx, ext, db_range=50, gain=30, figsize=(4,6))
    """
    if log_scale:
        Sxx_db = power2dB(Sxx, db_range, gain)
    else :
        Sxx_db = Sxx
    
    ax, fig = plot2d(Sxx_db, extent=extent, colorbar=colorbar, 
                     interpolation=interpolation, **kwargs)
    return ax

#%%
def rand_cmap(
    nlabels,
    type="bright",
    first_color_black=True,
    last_color_black=False,
    seed=321,
    verbose=False):
    """
    Creates a random colormap to be used together with matplotlib. Useful for segmentation tasks
    
    Parameters
    ----------
    nlabels : int
        Number of labels (size of colormap)
    type : string
        'bright' for strong colors, 'soft' for pastel colors. Default is 'bright'
    first_color_black : bool, optional
        Option to use first color as black. Default is True  
    last_color_black : bool, optional   
        Option to use last color as black. Default is False
    seed : int, optional
        Fix the seed of the random engine. Default is 321   
    verbose  : bool, optional   
        Prints the number of labels and shows the colormap. Default is False
    
    Returns
    -------
    random_colormap : Colormap 
        Colormap type used by matplotlib

    References
    ----------
    adapted from https://github.com/delestro/rand_cmap author : delestro    
    """

    # initialize the random seed in order to get always the same random order
    np.random.seed(seed=seed)

    if verbose:
        print("Number of labels: " + str(nlabels))

    # Generate color map for bright colors, based on hsv
    if type == "bright":
        randHSVcolors = [
            (
                np.random.uniform(low=0.1, high=1),
                np.random.uniform(low=0.2, high=1),
                np.random.uniform(low=0.5, high=1),
            )
            for i in range(nlabels)
        ]

        # Convert HSV list to RGB
        randRGBcolors = []
        for HSVcolor in randHSVcolors:
            randRGBcolors.append(
                colorsys.hsv_to_rgb(HSVcolor[0], HSVcolor[1], HSVcolor[2])
            )

    # Generate soft pastel colors, by limiting the RGB spectrum
    if type == "soft":
        low = 0.6
        high = 0.95
        randRGBcolors = [
            (
                np.random.uniform(low=low, high=high),
                np.random.uniform(low=low, high=high),
                np.random.uniform(low=low, high=high),
            )
            for i in range(nlabels)
        ]

    if first_color_black:
        randRGBcolors[0] = [0, 0, 0]

    if last_color_black:
        randRGBcolors[-1] = [0, 0, 0]

    random_colormap = LinearSegmentedColormap.from_list(
        "new_map", randRGBcolors, N=nlabels
    )

    # Display colorbar
    if verbose:
        from matplotlib import colors, colorbar
        from matplotlib import pyplot as plt

        fig, ax = plt.subplots(1, 1, figsize=(15, 0.5))

        bounds = np.linspace(0, nlabels, nlabels + 1)
        norm = colors.BoundaryNorm(bounds, nlabels)

        colorbar.ColorbarBase(
            ax,
            cmap=random_colormap,
            norm=norm,
            spacing="proportional",
            ticks=None,
            boundaries=bounds,
            format="%1i",
            orientation=u"horizontal",
        )

    return random_colormap


#%%
def crop_image(im, tn, fn, fcrop=None, tcrop=None):
    """
    Crop a spectrogram (or an image) in time (horizontal X axis) and frequency
    (vertical y axis)
    
    Parameters
    ----------
    im : 2d ndarray
        image to be cropped
    
    tn, fn : 1d ndarray
        tn is the time vector. fn is the frequency vector. They are required 
        in order to know the correspondance between pixels and (time,frequency)
    
    fcrop, tcrop : list of 2 scalars [min, max], optional, default is None
        fcrop corresponds to the min and max boundary frequency values
        tcrop corresponds to the min and max boundary time values
                
    Returns
    -------
    im : 2d ndarray
        image cropped
        
    tn, fn, 1d ndarray
        new time and frequency vectors
    
    Examples
    --------
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> Sxx_power,tn,fn,_ = maad.sound.spectrogram(w,fs)
    >>> Lxx = maad.spl.power2dBSPL(Sxx_power, gain=42) # convert into dB SPL
    >>> fig_kwargs = {'vmax': Lxx.max(),
                    'vmin':0,
                    'extent':(tn[0], tn[-1], fn[0], fn[-1]),
                    'title':'Power spectrogram density (PSD)',
                    'xlabel':'Time [s]',
                    'ylabel':'Frequency [Hz]',
                    }
    >>> fig, ax = maad.util.plot2d(Lxx,**fig_kwargs)      
    >>> Lxx_crop, tn_crop, fn_crop = maad.util.crop_image(Lxx, tn, fn, fcrop=(2000,10000), tcrop=(0,30))
    >>> fig_kwargs = {'vmax': Lxx.max(),
                    'vmin':0,
                    'extent':(tn_crop[0], tn_crop[-1], fn_crop[0], fn_crop[-1]),
                    'title':'Crop of the power spectrogram density (PSD)',
                    'xlabel':'Time [s]',
                    'ylabel':'Frequency [Hz]',
                    }
    >>> fig, ax = maad.util.plot2d(Lxx_crop,**fig_kwargs)  
    """

    if tcrop is not None:
        indt = (tn >= tcrop[0]) * (tn <= tcrop[1])
        im = im[:, indt]
        # redefine tn
        tn = tn[np.where(indt > 0)]
    if fcrop is not None:
        indf = (fn >= fcrop[0]) * (fn <= fcrop[1])
        im = im[
            indf,
        ]
        # redefine fn
        fn = fn[np.where(indf > 0)]

    return im, tn, fn


#%%
def save_figlist(fname, figlist):
    """
    Save a list of figures or spectrograms to disk.
    
    Parameters
    ----------
    fname: string
        Suffix string add to the filename. The extension should be specified since it 
        indicates the image format.

    Notes
    -----
    This function does not return any variable.
        
    """
    for i, fig in enumerate(figlist):
        fname_save = "%d_%s" % (i, fname)
        imsave(fname_save, fig)

#%%
def plot_features_map(df, norm=True, mode="24h", **kwargs):
    """
    Plot features values on a heatmap.
    
    The plot has the features the vertical axis and the time on the horizontal axis.
    
    Parameters
    ----------
    df : Panda DataFrame
        DataFrame with features (ie. indices).
    
    norm : boolean, default is True
        if True, the features are scaled between 0 to 1
        
    mode : string in {'24h'}, default is '24h'
        Select if the timeline of the phenology:
            
            -'24h' : average of the results over a day
            - otherwise, the timeline is the timeline of the dataframe
            
    **kwargs
        Specific to this function:
        
        - ftime : Time format to display as x label by default '%Y-%m-%d'(see https://docs.python.org/fr/3.6/library/datetime.html?highlight=strftime#strftime-strptime-behavior)
            
        Specific to matplotlib:
            
        - figsize : tuple of integers, optional, default: (4,10) width, height in inches.  
        
        - title : string, optional, default : 'Spectrogram' title of the figure
        
        - xlabel : string, optional, default : 'Time [s]' label of the horizontal axis

        - ylabel : string, optional, default : 'Amplitude [AU]' label of the vertical axis
            
        - xticks : tuple of ndarrays, optional, default : none
            * ticks : array_like => A list of positions at which ticks should be placed. You can pass an empty list to disable yticks.
            * labels : array_like, optional =>  A list of explicit labels to place at the given locs.
            
        - yticks : tuple of ndarrays, optional, default : none
            * ticks : array_like => A list of positions at which ticks should be placed. You can pass an empty list to disable yticks.
            * labels : array_like, optional =>  A list of explicit labels to place at the given locs.
        
        - cmap : string or Colormap object, optional, default is 'gray'
            See https://matplotlib.org/examples/color/colormaps_reference.html
            in order to get all the  existing colormaps
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic', 
            'viridis'...
        
        - vmin, vmax : scalar, optional, default: None
            `vmin` and `vmax` are used in conjunction with norm to normalize
            luminance data.  Note if you pass a `norm` instance, your
            settings for `vmin` and `vmax` will be ignored.
        
        - extent : list of scalars [left, right, bottom, top], optional, default: None
            The location, in data-coordinates, of the lower-left and
            upper-right corners. If `None`, the image is positioned such that
            the pixel centers fall on zero-based (row, column) indices.
        
        - now : boolean, optional, default : True
            if True, display now. Cannot display multiple images. 
            To display mutliple images, set now=False until the last call for 
            the last image      
            
        ... and more, see matplotlib
        
    Returns
    -------
    fig : Figure
        The Figure instance 
    ax : Axis
        The Axis instance   
    
    Examples
    --------
    see plot_extract_alpha_indices.py advanced example for a complete example
    
    >>> import numpy as np
    >>> import pandas as pd
    >>> np.random.seed(2021)
    >>> M = np.random.rand(24, 7)
    >>> df = pd.DataFrame(M)
    >>> indices = ['A','B','C','D','E','F','G']
    >>> df.columns = indices
    >>> df.index =pd.date_range(start=pd.Timestamp('00:00:00'), end=pd.Timestamp('23:00:00'), freq='1H')
    >>> maad.util.plot_features_map(df) 
    """

    if isinstance(df, pd.DataFrame) == False:
        raise TypeError("df must be a Pandas Dataframe")
    elif isinstance(df.index, pd.DatetimeIndex) == False:
        raise TypeError("df must have an index of type DateTimeIndex")

    if mode == "24h":
        # Mean values by hour
        df = df.groupby(df.index.hour).mean()
        # Get the list of unique index of type 'hour'
        x_label = [i + j for i, j in zip(map(str, df.index.values), ["h"] * len(df))]
    else:
        # Get the list of unique index of type 'hour'
        ftime = kwargs.pop("ftime", "%Y-%m-%d")
        x_value = df.index.strftime(ftime)
        x_label = x_value.tolist()

    if norm:
        df = linear_scale(df)

    # kwargs
    cmap = kwargs.pop("cmap", "RdBu_r")
    figsize = kwargs.pop("figsize", None)

    # plot
    if figsize is None :
        fig = plt.figure(figsize=(len(df)*0.33, len(list(df))*0.27))
    ax = fig.add_subplot(111)
    caxes = ax.matshow(df.transpose(), cmap=cmap, aspect="auto", **kwargs)
    if norm :
        fig.colorbar(caxes, shrink=0.75, label="Normalized value")
    else :
        fig.colorbar(caxes, shrink=0.75, label="Value")    
    # Set ticks on both sides of axes on
    ax.tick_params(axis="x", bottom=True, top=False, labelbottom=True, labeltop=False)
    # We want to show all ticks...
    ax.set_yticks(np.arange(len(df.columns)))
    ax.set_xticks(np.arange(len(x_label)))
    # ... and label them with the respective list entries
    ax.set_yticklabels(df.columns)
    ax.set_xticklabels(x_label)
    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_yticklabels(), rotation=0, ha="right", fontsize=10)
    plt.setp(ax.get_xticklabels(), rotation=90, ha="center", fontsize=10)
    fig.tight_layout()
    plt.show()

    return fig, ax

#%%
def heatmap_by_date_and_time (
                    dataframe,
                    disp_column,
                    date_format = "%V", # see https://docs.python.org/3/library/datetime.html 
                    date_range = [1,53],
                    time_resolution = "30T",
                    time_range = ["00:00", "23:59"],
                    start_hour = "00:00",
                    full_display = False,
                    date_min_to_disp = 1,
                    date_max_to_disp = 53,
                    cb_legend = "",
                    display = True,
                    verbose = False,
                    **kwargs,
                    ) :
    """
    Plot a heatmap of a features by time (x-axis) and date (y-axis).

    Parameters
    ----------
    dataframe : pandas.DataFrame
        The input DataFrame containing the data.
        Must contain a column (or index) date in the format %Y-%m-%d %H:%M%S
    disp_column : str
        The name of the column to be displayed.
    date_format : str, optional
        The format of the date. The default is "%V".
        (See https://docs.python.org/3/library/datetime.html for format codes)
        Possible format are :
        - "%V" for week number (from 1 to 53)
        - "%m" for Month (from 1 to 12)
        - "%d" for Day (from 1 to 31 depending on the month)
        - "%m-%d" for Date without year (from 01-01 to 12-31)
        - "%y-%m-%d" for Date with year 
    date_range : list of int, optional
        The range of date types to include. The default is [1, 53].
        The format of the range depends on date_format.
        For instance, to get all the samples in the March, date_range would be 
        [03-01, 03-31] and date_format would be "%m-%d"
    time_resolution : str, optional
        The time resolution. The default is "30T".
        T is for minute. 30T means a time resolution of 30 min. 
        Everything within this intervale is averaged
        Other time formats are H for hour and D for day.
    time_range : list of str, optional
        The time range to consider. The default is ["00:00", "23:59"].
    start_hour : str, optional
        The start hour. The default is "12:00".
    full_display : bool, optional
        Whether to display the full date range. The default is False.
    date_min_to_disp : int, optional
        The minimum date on the y-axis. The default is 1.
        The value depends on the date_format. 
        See date_format to know the possible formats and the range of possible values
    date_max_to_disp : int, optional
        The maximum date on the y-axis. The default is 53.
        The value depends on the date_format. 
        See date_format to know the possible formats and the range of possible values
    cb_legend : str, optional
        The colorbar legend label. The default is "".
    verbose : bool, optional
        If True, display verbose information. The default is False.
    **kwargs
        Specific to this function:
        
        - ftime : Time format to display as x label by default '%Y-%m-%d'(see https://docs.python.org/fr/3.6/library/datetime.html?highlight=strftime#strftime-strptime-behavior)
            
        Specific to matplotlib:
            
        - figsize : tuple of integers, optional, default: (4,10) width, height in inches.  
        
        - title : string, optional, default : 'Spectrogram' title of the figure
        
        - xlabel : string, optional, default : 'Time [s]' label of the horizontal axis

        - ylabel : string, optional, default : 'Amplitude [AU]' label of the vertical axis
            
        - xticks : tuple of ndarrays, optional, default : none
            * ticks : array_like => A list of positions at which ticks should be placed. You can pass an empty list to disable yticks.
            * labels : array_like, optional =>  A list of explicit labels to place at the given locs.
            
        - yticks : tuple of ndarrays, optional, default : none
            * ticks : array_like => A list of positions at which ticks should be placed. You can pass an empty list to disable yticks.
            * labels : array_like, optional =>  A list of explicit labels to place at the given locs.
        
        - cmap : string or Colormap object, optional, default is 'gray'
            See https://matplotlib.org/examples/color/colormaps_reference.html
            in order to get all the  existing colormaps
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic', 
            'viridis'...
        
        - vmin, vmax : scalar, optional, default: None
            `vmin` and `vmax` are used in conjunction with norm to normalize
            luminance data.  Note if you pass a `norm` instance, your
            settings for `vmin` and `vmax` will be ignored.
        
        - extent : list of scalars [left, right, bottom, top], optional, default: None
            The location, in data-coordinates, of the lower-left and
            upper-right corners. If `None`, the image is positioned such that
            the pixel centers fall on zero-based (row, column) indices.
        
        - now : boolean, optional, default : True
            if True, display now. Cannot display multiple images. 
            To display mutliple images, set now=False until the last call for 
            the last image      
            
        ... and more, see matplotlib

    Returns
    -------
    df_mean : pd.DataFrame
        The heatmap with mean values.
    
    df_std :pd.DataFrame
        The heatmap with standard deviation values.

    fig : matplotlib.figure.Figure
        The generated matplotlib figure.

    ax : matplotlib.axes.Axes
        The generated matplotlib axis.
    """

    # test if dataframe is not a pd.DataFrame
    if isinstance(dataframe, pd.DataFrame) == False:
        if verbose :
            print("***WARNING*** dataframe must be a valid pandas dataframe")
        return
    
    # copy the dataframe
    df = dataframe.copy()

    # keep some columns
    df = df.filter([disp_column,'date'])

    try :
        if not('date' in df)  :  
            df = df.reset_index(['date'])
        # convert date to datetime
        df['date'] = pd.to_datetime(df['date'])
        # set the index to date
        df.set_index(['date'], inplace=True)
        # convert index to datetime 
        df.index = pd.DatetimeIndex(df.index)
        # sort dataframe by date
        df = df.sort_index(axis=0)
    except :
        raise Exception("***WARNING*** df must have a valid date index that could be converted in Datetime type")

    # filter by time
    df = _filter_dataframe_by_datetime(df, "%H:%M", time_range)
    # filter by date_type
    df = _filter_dataframe_by_datetime(df, date_format, date_range)

    # # Extract hour and minute
    df['time'] = df.index.strftime("%H:%M")

    # # Extract week, year, weekday or date
    df['date_type'] = df.index.strftime(date_format) 

    # create a matrix with date_type (week, year, weekday, date) as columns and time (hour) as rows
    df_mean = df.groupby(['time', 'date_type'])[disp_column].aggregate('mean').unstack()

    # Resample the timeline. Need to reformat the time in datetime before resampling
    df_mean = df_mean.reset_index()
    df_mean.index = pd.to_datetime(df_mean['time'], format='%H:%M')
    df_mean.drop(columns=['time'], inplace = True)
    df_mean = df_mean.resample(time_resolution).mean()
    df_std = df_mean.resample(time_resolution).std()
    df_mean.index = df_mean.index.strftime("%H:%M")
    df_std.index = df_std.index.strftime("%H:%M")

    if display == True : 

        # try to use seaborn if installed
        try:
            import seaborn as sns
            sns.set_theme("paper")
        except ImportError as e:
            if verbose :
                print("seaborn is not installed, use default settings")

        if full_display == True :
            # Create a list of unique index of type 'hour' for x axis
            time_min = datetime.strptime("00:00", "%H:%M")
            time_max = datetime.strptime("23:59", "%H:%M")
            x_label = pd.date_range(time_min, time_max, freq=time_resolution).strftime("%H:%M")
            x_label = x_label.values
            # Create a vector for y axis depending on date_format
            if date_format == "%V" :
                y_label = [f'{x:02d}' for x in np.arange(date_min_to_disp,date_max_to_disp+1)]
            elif date_format == "%m" :
                y_label = [f'{x:02d}' for x in np.arange(date_min_to_disp,date_max_to_disp+1)]   
            elif (date_format == "%y-%m-%d") or (date_format == "%m-%d") or (date_format == "%d"):
                date_min_to_disp = datetime.strptime(str(date_min_to_disp), date_format)
                date_max_to_disp = datetime.strptime(str(date_max_to_disp), date_format)
                y_label = pd.date_range(date_min_to_disp, date_max_to_disp, freq='D').strftime(date_format)
                y_label = y_label.values
            else:
                # by default per week
                y_label = [f'{x:02d}' for x in np.arange(1,53)]
            # create an prefilled dataframe with all weeks and time
            df2 = pd.DataFrame(index=x_label, 
                            columns=y_label,
                            dtype = 'float')
            df2.reset_index(inplace=True)
            df2 = df2.rename(columns={"index":"time"})
            df2.set_index("time", inplace=True)

            for idx, row in df_mean.iterrows() :
                df2.loc[idx] = row
            
            df_mean=df2
        
        # rename the unique column with the corresponding name
        if date_format == "%V" : df_mean.columns.name='Week number'
        if date_format == "%m" : df_mean.columns.name='Month'
        if date_format == "%d" : df_mean.columns.name='Day'
        if date_format == "%m-%d" : df_mean.columns.name='Date'
        if date_format == "%y-%m-%d" : df_mean.columns.name='Date'

        # shift the time to start at start_hour and finish at start_hour, in order to center
        # the graph at start_hour + 12h
        df_part1 = df_mean[df_mean.index >= start_hour]
        df_part2 = df_mean[df_mean.index < start_hour]
        df_part2.sort_index(inplace=True)
        df = pd.concat([df_part1,df_part2])  
        
        # Get the list of unique index of type 'hour'
        x_label = [i + j for i, j in zip(map(str, df.index.values), [" "] * len(df))]

        # plot
        # matplotlib parameters
        fig = kwargs.pop("fig", None)
        ax = kwargs.pop("ax", None) 
        xlabel = kwargs.pop("xlabel", "Hours")
        ylabel = kwargs.pop("ylabel", df_mean.columns.name)
        now = kwargs.pop("now", True)
        vmin = kwargs.pop("vmin", np.nanpercentile(df,1))
        vmax = kwargs.pop("vmax", np.nanpercentile(df,99))
        cmap = kwargs.pop("cmap", "RdPu")
        title = kwargs.pop("title", None)
        figsize = kwargs.pop("figsize", (len(df)*0.33*0.75, len(list(df))*0.26*0.75 +0.75))
        
        # Create a figure and a subplot associated
        if (fig is None) or (ax is None) :
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111)

        # display image
        caxes = ax.matshow(df.transpose(), 
                        interpolation = "None",
                        aspect="auto", 
                        cmap = cmap,
                        vmin = vmin,
                        vmax = vmax,
                        **kwargs
                        )
        if cb_legend =="" :
            fig.colorbar(caxes, shrink=0.75, label=disp_column)
        else:
            fig.colorbar(caxes, shrink=0.75, label=cb_legend)
        # Set ticks on both sides of axes on
        ax.tick_params(which='major', axis="x", bottom=False, top=False, 
                    labelbottom=True, labeltop=False)
        ax.tick_params(which='minor', axis="x", length = 0)
        # We want to show all ticks...
        ax.set_yticks(np.arange(len(df.columns)))
        ax.set_xticks(np.arange(len(x_label)))
        # ... and label them with the respective list entries
        ax.set_yticklabels(df.columns)
        ax.set_xticklabels(x_label)
        
        # Minor ticks
        ax.set_xticks(np.arange(-0.5, len(df.index), 1), minor=True)
        ax.set_yticks(np.arange(-0.5, len(list(df)), 1), minor=True)

        # Gridlines based on minor ticks
        ax.grid(which='major', color='w', linestyle='-', linewidth=0)
        ax.grid(which='minor', color='w', linestyle='-', linewidth=1)
        
        # add title to the axis
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
                
        # Rotate the tick labels and set their alignment.
        plt.setp(ax.get_yticklabels(), rotation=0, ha="right", fontsize=10)
        plt.setp(ax.get_xticklabels(), rotation=90, ha="center", fontsize=10)
        plt.tight_layout()

        # Adding a title to the figure
        if title is not None:
            fig.suptitle(title)

        if now :
            plt.show()
    else :
        fig = None
        ax = None

    return df_mean, df_std, fig, ax

# =============================================================================

def plot_features(df, ax=None, norm=True, mode="24h", **kwargs):
    """
    Plot the variation of features values (ie. indices) in the DataFrame obtained 
    with ``maad.features``.
            
    Parameters
    ----------
    df : Panda DataFrame
        DataFrame with features (ie. indices).
    
    norm : boolean, default is True
        if True, the features are normalized by the max
        
    mode : string in {'24h'}, default is '24h'
        Select if the timeline of the phenology :
            -'24h' : average of the results over a day
            - otherwise, the timeline is the timeline of the dataframe
            
    **kwargs
        - figsize : tuple of integers, optional, default: (4,10)
            width, height in inches.  
        
        - figtitle : string, optional, default : ''
            title of the figure
        
        - xlabel : string, optional, default : 'Time [s]'
            label of the horizontal axis
        
        - ylabel : string, optional, default : 'Amplitude [AU]'
            label of the vertical axis
            
        - xticks : tuple of ndarrays, optional, default : none
            * ticks : array_like => A list of positions at which ticks should be placed. You can pass an empty list to disable yticks.
            * labels : array_like, optional =>  A list of explicit labels to place at the given locs.
            
        - yticks : tuple of ndarrays, optional, default : none
            * ticks : array_like => A list of positions at which ticks should be placed. You can pass an empty list to disable yticks.
            * labels : array_like, optional =>  A list of explicit labels to place at the given locs.

        - now : boolean, optional, default : True
            if True, display now. Cannot display multiple images. 
            To display mutliple images, set now=False until the last call for 
            the last image      
            
        ... and more, see matplotlib
        
    Returns
    -------
    fig : Figure
        The Figure instance 
    ax : Axis
        The Axis instance
        
    Examples
    --------
    see plot_extract_alpha_indices.py advanced example for a complete example
    
    >>> import numpy as np
    >>> import pandas as pd
    >>> np.random.seed(2021)
    >>> M = np.random.rand(24, 2)
    >>> df = pd.DataFrame(M)
    >>> indices = ['A','B']
    >>> df.columns = indices
    >>> df.index =pd.date_range(start=pd.Timestamp('00:00:00'), end=pd.Timestamp('23:00:00'), freq='1H')
    >>> maad.util.plot_features(df) 
        
    """
    if isinstance(df, pd.DataFrame) == False:
        raise TypeError("df must be a Pandas Dataframe")
    elif isinstance(df.index, pd.DatetimeIndex) == False:
        raise TypeError("df must have an index of type DateTimeIndex")

    if mode == "24h":
        # Mean values by hour
        df = df.groupby(df.index.hour).mean()

    if norm:
        df = linear_scale(df)

    # plot
    import itertools
    from matplotlib.lines import Line2D

    list_markers = tuple(list(Line2D.markers.keys())[0:-4])
    markers = itertools.cycle(list_markers)
    prop_cycle = plt.rcParams["axes.prop_cycle"]
    colors = itertools.cycle(prop_cycle.by_key()["color"])

    figsize = kwargs.pop("figsize", (5, 5))
    kwargs.pop("label", None)
    now = kwargs.pop("now", True)

    # if no ax, create a figure and a subplot associated a figure otherwise
    # find the figure that belongs to ax
    if ax is None:
        fig, ax = plt.subplots(**kwargs)
        fig.set_size_inches(figsize)
    else:
        fig = ax.get_figure()

    for indice in list(df):
        ax.plot(
            df.index,
            df[indice],
            marker=next(markers),
            color=next(colors),
            linestyle="-",
            label=indice,
            **kwargs
        )

    if mode == "24h":
        ax.set_xlabel("Day time (Hour)")
    ax.grid()
    ax.legend()
    fig.tight_layout()

    # Display the figure now
    if now:
        plt.show()

    return fig, ax


# =============================================================================
def plot_correlation_map(df, R_threshold=0.75, method="spearman", **kwargs):
    """
    Plot the correlation map between indices in the DataFrame obtained
    with ``maad``.
        
    Parameters
    ----------
    df : Panda DataFrame
        DataFrame with features (ie. indices).
        
    R_threshold : scalar between 0 to 1, default is 0.75
        Show correlations with R higher than R_threshold
        
    method : string in {'spearman', 'pearson'}, default is 'spearman'
        Choose the correlation type
    
    **kwargs
        - figsize : tuple of integers, optional, default: (4,10)
            width, height in inches.  
        
        - title : string, optional, default : 'Spectrogram'
            title of the figure
        
        - xlabel : string, optional, default : 'Time [s]'
            label of the horizontal axis
        
        - ylabel : string, optional, default : 'Amplitude [AU]'
            label of the vertical axis
            
        - xticks : tuple of ndarrays, optional, default : none
            * ticks : array_like => A list of positions at which ticks should be placed. You can pass an empty list to disable yticks.
            * labels : array_like, optional =>  A list of explicit labels to place at the given locs.
            
        - yticks : tuple of ndarrays, optional, default : none
            * ticks : array_like => A list of positions at which ticks should be placed. You can pass an empty list to disable yticks.
            * labels : array_like, optional =>  A list of explicit labels to place at the given locs.
        
        - cmap : string or Colormap object, optional, default is 'gray'
            See https://matplotlib.org/examples/color/colormaps_reference.html
            in order to get all the  existing colormaps
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic', 
            'viridis'...
        
        - vmin, vmax : scalar, optional, default: None
            `vmin` and `vmax` are used in conjunction with norm to normalize
            luminance data.  Note if you pass a `norm` instance, your
            settings for `vmin` and `vmax` will be ignored.
        
        - extent : list of scalars [left, right, bottom, top], optional, default: None
            The location, in data-coordinates, of the lower-left and
            upper-right corners. If `None`, the image is positioned such that
            the pixel centers fall on zero-based (row, column) indices.
        
        - now : boolean, optional, default : True
            if True, display now. Cannot display multiple images. 
            To display mutliple images, set now=False until the last call for 
            the last image      
            
        ... and more, see matplotlib
        
    Returns
    -------
    fig : Figure
        The Figure instance 
    ax : Axis
        The Axis instance
        
    Examples
    --------
    see plot_extract_alpha_indices.py advanced example for a complete example
    
    >>> import numpy as np
    >>> import pandas as pd
    >>> np.random.seed(2021)
    >>> M = np.random.rand(10, 10)
    >>> df = pd.DataFrame(M)
    >>> indices = ['A','B','C','D','E','F','G','H','I','J']
    >>> df.columns = indices
    >>> maad.util.plot_correlation_map(df, R_threshold=0)
    """
    # Correlation matrix with only the columns that contain numeric values
    numeric_columns = df.select_dtypes(include='number')
    corr_matrix = numeric_columns.corr()

    # pop kwargs
    figsize = kwargs.pop("figsize", (10, 8))
    cmap = kwargs.pop("cmap", "RdBu_r")
    label_colorbar = kwargs.pop("label_colorbar", "R")

    fig = plt.figure()
    fig.set_size_inches(figsize)
    ax = fig.add_subplot(111)
    caxes = ax.matshow(corr_matrix[abs(corr_matrix) ** 2 > R_threshold ** 2], cmap=cmap)
    fig.colorbar(caxes, shrink=0.75, label=label_colorbar)

    # We want to show all ticks...
    ax.set_xticks(np.arange(len(corr_matrix.columns)))
    ax.set_yticks(np.arange(len(corr_matrix.columns)))
    # ... and label them with the respective list entries
    ax.set_xticklabels(corr_matrix.columns)
    ax.set_yticklabels(corr_matrix.columns)

    # Rotate the tick labels, set their alignment and fontsize
    plt.setp(ax.get_xticklabels(), rotation=90, ha="center", fontsize=7)
    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_yticklabels(), rotation=0, ha="right", fontsize=8)

    fig.tight_layout()
    plt.show()

    return fig, ax


# =============================================================================
def false_Color_Spectro(
    df,
    indices=None,
    plim=(1, 99),
    reverseLUT=False,
    permut=False,
    unit="minutes",
    verbose=False,
    display=False,
    savefig=None,
    **kwargs
):
    """
    Create False Color Spectrogram from indices obtained by MAAD.
    Only indices than can be computed bin per bin (ie frequency per frequency)
    are used to create False Color Spectro. They are called xxx_per_bin.
        
    Parameters
    ----------
    df : Panda DataFrame
        DataFrame with indices per frequency bin.
        
    indices : list, default is None
        List of indices. 
        If permut is False (see permut), if indices is None : 1st indices is red (R), 2nd indice is green (G) 
        and the 3rd indices is blue (B).
        if indices is a list of 3 indices or more, only the 3 first indices (triplet) are used to create the false color spectro.
        if permut is True, if indices is None : All indices in df are used
        if indices is a list of 3 indices or more, all the indices in the 
        list are used to create the false color spectro with all the possibile permutations.
    
    plim : list of 2 scalars, default is (1,99)
        Set the minimum and maximum percentile to define the min and max value 
        of each indice. These values are then used to clip the values of each 
        indices between these limits.
             
    reverseLUT: boolean, default is False
        LUT : look-up table.
        if False, the highest indice value is the brigthest color (255) 
        and the lowest indice value is the darkest (0)
        if True, it's the reverse order, the highest indice value is the 
        darkest color (0) and the lowest is the brighest (255)     
    
    permut : boolean, default is False
        if True, all the permutations possible for red, green, blue are computed
        among the list of indices (see indices)
    
    unit : string in {'minutes', 'hours', 'days', 'weeks'}, default is 'minutes'
        Select the unit base for x axis
        
    verbose : boolean, default is False
        print indices on the default terminal
        
    display : boolean, default is False
        Display False Color Spectro
        
    savefig : string, optional, default is None 
        if not None, figures will be safe. Savefig is the prefix of the save
        filename.
    
    kwargs, optional
       - dpi : scalar, optional, default 96
       - format : str, optional, default .png
    
    Returns
    -------
    false_color_image : ndarray of scalars
        Matrix with ndim = 4 if multiple false color spectro or ndim = 3, if
        single false color spectro with 3 colors : R, G, B
            
    triplet : list
        List of triplet of indices corresponding to each false color spectro
        
    Examples
    --------
    see plot_extract_alpha_indices.py advanced example for a complete example 
    
    >>> import numpy as np
    >>> import pandas as pd
    >>> A_per_bin = [[0.5,0.4,0.3,0.2,0.1],
                     [0.3,0.3,0.3,0.3,0.3],
                     [0.5,0.5,0.3,0.0,0.0],
                     [0.2,0.2,0.3,0.0,0.0],
                     [0.0,0.0,0.3,0.0,0.0]]
    >>> B_per_bin = [[0.5,0.5,0.5,0.5,0.5],
                     [0.5,0.5,0.5,0.5,0.5],
                     [0.5,0.5,0.2,0.0,0.0],
                     [0.5,0.2,0.1,0.0,0.0],
                     [0.5,0.2,0.0,0.0,0.0]]
    >>> C_per_bin = [[0.0,0.7,0.0,0.5,0.5],
                     [0.7,0.2,0.0,0.5,0.5],
                     [0.5,0.5,0.7,0.2,0.2],
                     [0.0,0.7,0.0,0.2,0.2],
                     [0.7,0.2,0.0,0.2,0.2]]   
    >>> df = pd.DataFrame([A_per_bin,B_per_bin,C_per_bin])
    >>> df = df.T
    >>> df.columns = ['A_per_bin','B_per_bin','C_per_bin']
    >>> df.index = pd.date_range('20200101', periods=len(df))
    >>> maad.util.false_Color_Spectro (df, display=True ,unit='days', figsize=[3,3])
    
    """
    # sort dataframe by date
    df = df.sort_index(axis=0)
    # set index as DatetimeIndex
    df = df.set_index(pd.DatetimeIndex(df.index))
    # remove column file
    if "file" in df.columns:
        df = df.drop(columns="file")
    # test if frequencies is in columns
    if "frequencies" in df.columns:
        # get frequencies and remove the column
        # test if type of df['frequencies'] is number or str
        if isinstance(df.frequencies.iloc[0], str):
            fn = df["frequencies"].apply(literal_eval).iloc[0]
        else:
            fn = df["frequencies"].iloc[0]
        # drop frequencies
        df = df.drop(columns="frequencies")
    else:
        fn = np.arange(0, len(df))

    # Set the list of indices if indices is None
    if indices is None:
        indices = list(df)

    # for each indice, check if values type is string and convert into scalars
    # Values are vectors of strings when data are loading from csv
    for indice in indices:
        # test if type of df[indice] is string
        if isinstance(df[indice].iloc[0], str):
            # convert string into scalars
            df[indice] = df[indice].apply(literal_eval)
            
    # create a dataframe with normalized values for each indices
    df_z = pd.DataFrame()

    for indice in indices:
        z = []
        if verbose:
            print(indice)
        for v in df[indice]:
            z.append(v)
        z = np.asarray(z).T
        # Select the min and max value for each indice
        z_min = np.percentile(z, plim[0])
        z_max = np.percentile(z, plim[1])
        # clip the value to the min and max found
        z = np.clip(z, z_min, z_max)
        # linear conversion
        if reverseLUT == True:
            # between 1 to 0
            z = linear_scale(z, 1, 0)
        else:
            # between 0 to 1
            z = linear_scale(z, 0, 1)

        df_z[indice] = [z * 255]

    # find all permutation of 3 indices among all indices
    if permut == True:
        import itertools

        per = itertools.permutations(list(df_z), 3)
        triplet = []
        for val in per:
            triplet.append([*val])
    else:
        triplet = []
        triplet.append(indices[0])
        triplet.append(indices[1])
        triplet.append(indices[2])
        triplet = [triplet]

    #####################

    # get the number of pixels along frequency (Nf) and time (Nt)
    Nf, Nt = df_z[triplet[0][0]].values.tolist()[0].shape

    # test if figsize is in kwargs
    figsize = kwargs.pop("figsize", None)
    # if figsize is not in kwargs
    if figsize is None:
        figsize = (6 * Nt / 250, 10 * Nf / 512)

    fig_kwargs = {"figsize": figsize, "tight_layout": "tight_layout"}

    # number of days in the period
    deltaT = df.index.max() - df.index.min()

    # unit
    if unit == "minutes":
        normT = 60e9
        xlabel = "Minutes"
    elif unit == "hours":
        normT = 60e9 * 60
        xlabel = "Hours"
    elif unit == "days":
        normT = 60e9 * 60 * 24
        xlabel = "Days"
    elif unit == "weeks":
        normT = 60e9 * 60 * 24 * 7
        xlabel = "Weeks"
    else:
        normT = 60e9
        xlabel = "Minutes"

    false_color_image = []
    for tt in np.arange(len(triplet)):

        # create the false color image (R,G,B)
        z0 = df_z[triplet[tt][0]].values.tolist()[0]
        z1 = df_z[triplet[tt][1]].values.tolist()[0]
        z2 = df_z[triplet[tt][2]].values.tolist()[0]
        false_color_image.append((np.dstack((z0, z1, z2))).astype(np.uint8))

        # Display the False Color Spectro
        if display:
            plt.rcParams.update({"font.size": 10})
            plt.rcParams.update({"font.family": "serif"})
            fig = plt.figure(facecolor="white", **fig_kwargs)
            plt.imshow(
                false_color_image[tt],
                aspect="auto",
                origin="lower",
                extent=(fn[0], deltaT.value / normT, 0, fn[-1]),
                **kwargs
            )
            plt.xlabel(xlabel)
            plt.ylabel("Frequency (Hz)")
            plt.title(
                "False Color Spectro "
                + "\n"
                + " [R:"
                + triplet[tt][0][:-8]
                + "; "
                + "G:"
                + triplet[tt][1][:-8]
                + "; "
                + "B:"
                + triplet[tt][2][:-8]
                + "]",
                size=12,
            )
            plt.xticks(fontsize=10)
            plt.yticks(fontsize=10)
            plt.tight_layout()

            # SAVE FIGURE
            if savefig is not None:
                dpi = kwargs.pop("dpi", 96)
                bbox_inches = "tight"
                format = kwargs.pop("format", "png")
                filename = (
                    "_fcs_"
                    + triplet[tt][0][:-8]
                    + "_"
                    + triplet[tt][1][:-8]
                    + "_"
                    + triplet[tt][2][:-8]
                )
                full_filename = savefig + filename + "." + format
                if verbose:
                    print("\n" "save figure : %s" % full_filename)
                # save fig
                fig.savefig(
                    fname=full_filename,
                    dpi=dpi,
                    bbox_inches=bbox_inches,
                    format=format,
                    **kwargs
                )
                # close fig
                plt.close(fig)

    # convert into ndarray
    false_color_image = np.asarray(false_color_image)

    # test if there is only one False Color Spectro, then, remove the 1st dim
    if false_color_image.shape[0] == 1:
        false_color_image = false_color_image[0]
        triplet = triplet[0]

    return false_color_image, triplet
